#
# Copyright 2015-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import argparse
import numbers

import numpy as np
from numpy.polynomial import Polynomial

from .matplotlib_qt import set_window_geometry
from ..display.pause_debugplot import pause_debugplot


def polfit_residuals(
        x, y, deg, reject=None,
        color='b', size=75,
        xlim=None, ylim=None,
        xlabel=None, ylabel=None, title=None,
        use_r=False,
        show=True,
        geometry=(0, 0, 640, 480),
        debugplot=0,
        pdffile=None,
        fig=None
):
    """Polynomial fit with display of residuals and additional work with R.

    Parameters
    ----------
    x : 1d numpy array, float
        X coordinates of the data being fitted.
    y : 1d numpy array, float
        Y coordinates of the data being fitted.
    deg : int
        Degree of the fitting polynomial.
    reject : None or 1d numpy array (bool)
        If not None, it must be a boolean array indicating whether a
        particular point is rejected or not (i.e., the rejected points
        are flagged as True in this array). Rejected points are
        displayed but not used in the fit.
    color : single character or 1d numpy array of characters
        Color for all the symbols (single character) or for each
        individual symbol (array of color names with the same length as
        'x' or 'y'). If 'color' is a single character, the rejected
        points are displayed in red color, whereas when 'color' is an
        array of color names, rejected points are displayed with the
        color provided in this array.
    size : int
        Marker size for all the symbols (single character) or for each
        individual symbol (array of integers with the same length as
        'x' or 'y').
    xlim : tuple (floats)
        Plot limits in the X axis.
    ylim : tuple (floats)
        Plot limits in the Y axis.
    xlabel : string
        Character string for label in X axis.
    ylabel : string
        Character string for label in y axis.
    title : string
        Character string for graph title.
    use_r : bool
        If True, the function computes several fits, using R, to
        polynomials of degree deg, deg+1 and deg+2 (when possible).
    show : bool
        If True, the function shows the displayed image. Otherwise
        plt.show() is expected to be executed outside.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.
    pdffile : str or None
        Output PDF file name to save plot. This option is only
        employed when a plot has been generated.
    fig : `~matplotlib.figure.Figure`
        If not None, use this object to generate plot.

    Return
    ------
    poly : instance of Polynomial (numpy)
        Result from the polynomial fit using numpy Polynomial. Only
        points not flagged as rejected are employed in the fit.
    yres : 1d numpy array, float
        Residuals from polynomial fit. Note that the residuals are
        computed for all the points, including the rejected ones. In
        this way the dimension of this array is the same as the
        dimensions of the input 'x' and 'y' arrays.

    """

    # protections
    if type(x) is not np.ndarray:
        raise ValueError("x=" + str(x) + " must be a numpy.ndarray")
    elif x.ndim != 1:
        raise ValueError("x.ndim=" + str(x.ndim) + " must be 1")
    if type(y) is not np.ndarray:
        raise ValueError("y=" + str(y) + " must be a numpy.ndarray")
    elif y.ndim != 1:
        raise ValueError("y.ndim=" + str(y.ndim) + " must be 1")
    npoints = x.size
    if npoints != y.size:
        raise ValueError("x.size != y.size")
    if reject is not None:
        if npoints != reject.size:
            raise ValueError("x.size != reject.size")
    if not isinstance(deg, numbers.Integral):
        raise ValueError("deg=" + str(deg) +
                         " is not a valid integer")

    # select points for fit
    if reject is None:
        xfitted = np.copy(x)
        yfitted = np.copy(y)
        xrejected = None
        yrejected = None
        nfitted = npoints
        nrejected = 0
    else:
        xfitted = x[np.logical_not(reject)]
        yfitted = y[np.logical_not(reject)]
        xrejected = x[reject]
        yrejected = y[reject]
        # update number of points for fit
        nfitted = xfitted.size
        nrejected = sum(reject)

    if deg > nfitted - 1:
        raise ValueError("Insufficient nfitted=" + str(nfitted) +
                         " for deg=" + str(deg))

    # polynomial fits using R
    if use_r:
        from ..rutilities import LinearModelYvsX
        print("\n>>> Total number of points:", nfitted)
        # using orthogonal polynomials
        myfit = LinearModelYvsX(x=xfitted, y=yfitted, degree=deg, raw=False)
        print(">>> Fit with R, using orthogonal polynomials:")
        print(myfit.summary)
        pause_debugplot(debugplot)
        # fit using raw polynomials
        myfit = LinearModelYvsX(x=xfitted, y=yfitted, degree=deg, raw=True)
        print(">>> Fit with R, using raw polynomials:")
        print(myfit.summary)
        pause_debugplot(debugplot)

    # fit with requested degree (and raw polynomials)
    if deg == 0 and len(xfitted) == 1:  # constant (avoid fitting error)
        poly = Polynomial(yfitted[0])
    else:
        poly = Polynomial.fit(x=xfitted, y=yfitted, deg=deg)
    poly = Polynomial.cast(poly)

    # compute residuals
    yres = y - poly(x)  # of all the points
    yres_fitted = yfitted - poly(xfitted)  # points employed in the fit
    yres_rejected = None
    if nrejected > 0:
        yres_rejected = yrejected - poly(xrejected)  # points rejected

    if abs(debugplot) >= 10:
        print(">>> Polynomial fit:\n", poly)

    if abs(debugplot) % 10 != 0:
        # define colors, markers and sizes for symbols
        if np.array(color).size == 1:
            mycolor = np.array([color] * npoints)
            if reject is not None:
                mycolor[reject] = 'r'
        elif np.array(color).size == npoints:
            mycolor = np.copy(np.array(color))
        elif np.array(color).shape[0] == npoints:  # assume rgb color
            mycolor = np.copy(np.array(color))
        else:
            raise ValueError("color=" + str(color) +
                             " doesn't have the expected dimension")
        if np.array(size).size == 1:
            mysize = np.repeat([size], npoints)
        elif np.array(size).size == npoints:
            mysize = np.copy(np.array(size))
        else:
            raise ValueError("size=" + str(size) +
                             " doesn't have the expected dimension")

        if reject is None:
            cfitted = np.copy(mycolor)
            crejected = None
            sfitted = np.copy(mysize)
            srejected = None
        else:
            cfitted = mycolor[np.logical_not(reject)]
            crejected = mycolor[reject]
            sfitted = mysize[np.logical_not(reject)]
            srejected = mysize[reject]

        from numina.array.display.matplotlib_qt import plt
        if fig is None:
            fig = plt.figure()

        set_window_geometry(geometry)

        # residuals
        ax2 = fig.add_subplot(2, 1, 2)
        if xlabel is None:
            ax2.set_xlabel('x')
        else:
            ax2.set_xlabel(xlabel)
        ax2.set_ylabel('residuals')
        if xlim is None:
            xmin = min(x)
            xmax = max(x)
            dx = xmax - xmin
            if dx > 0:
                xmin -= dx/20
                xmax += dx/20
            else:
                xmin -= 0.5
                xmax += 0.5
        else:
            xmin, xmax = xlim
        ax2.set_xlim(xmin, xmax)
        ymin = min(yres_fitted)
        ymax = max(yres_fitted)
        dy = ymax - ymin
        if dy > 0:
            ymin -= dy/20
            ymax += dy/20
        else:
            ymin -= 0.5
            ymax += 0.5
        ax2.set_ylim(ymin, ymax)
        ax2.axhline(y=0.0, color="black", linestyle="dashed")
        ax2.scatter(xfitted, yres_fitted, color=cfitted,
                    marker='o',
                    edgecolor='k', s=sfitted)
        if nrejected > 0:
            ax2.scatter(xrejected, yres_rejected,
                        marker='x', s=srejected,
                        color=crejected)

        # original data and polynomial fit
        ax = fig.add_subplot(2, 1, 1, sharex=ax2)
        if ylabel is None:
            ax.set_ylabel('y')
        else:
            ax.set_ylabel(ylabel)
        ax.set_xlim(xmin, xmax)
        if ylim is None:
            ymin = min(y)
            ymax = max(y)
            dy = ymax - ymin
            if dy > 0:
                ymin -= dy/20
                ymax += dy/20
            else:
                ymin -= 0.5
                ymax += 0.5
        else:
            ymin, ymax = ylim
        ax.set_ylim(ymin, ymax)
        ax.scatter(xfitted, yfitted,
                   color=cfitted, marker='o', edgecolor='k',
                   s=sfitted, label="fitted data")
        xpol = np.linspace(start=xmin, stop=xmax, num=1000)
        ypol = poly(xpol)
        ax.plot(xpol, ypol, 'c-', label="fit")
        if nrejected > 0:
            ax.scatter(xrejected, yrejected,
                       marker='x', s=srejected, color=crejected,
                       label="rejected")

        # put a legend
        ax.legend(numpoints=1)

        # graph title
        if title is not None:
            plt.title(title)

        # save PDF file
        if pdffile is not None:
            plt.savefig(pdffile, pad_inches=0.02, bbox_inches='tight')

        if show:
            pause_debugplot(debugplot, pltshow=show, tight_layout=True)

    # return result
    return poly, yres


def polfit_residuals_with_sigma_rejection(
        x, y, deg, times_sigma_reject,
        color='b', size=75,
        xlim=None, ylim=None,
        xlabel=None, ylabel=None, title=None,
        use_r=None,
        show=True,
        geometry=(0, 0, 640, 480),
        debugplot=0,
        pdffile=None,
        fig=None
):
    """Polynomial fit with iterative rejection of points.

    This function makes use of function polfit_residuals for display
    purposes.

    Parameters
    ----------
    x : 1d numpy array, float
        X coordinates of the data being fitted.
    y : 1d numpy array, float
        Y coordinates of the data being fitted.
    deg : int
        Degree of the fitting polynomial.
    times_sigma_reject : float or None
        Number of times the standard deviation to reject points
        iteratively. If None, the fit does not reject any point.
    color : single character or 1d numpy array of characters
        Color for all the symbols (single character) or for each
        individual symbol (array of color names with the same length as
        'x' or 'y'). If 'color' is a single character, the rejected
        points are displayed in red color, whereas when 'color' is an
        array of color names, rejected points are displayed with the
        color provided in this array.
    size : int
        Marker size for all the symbols (single character) or for each
        individual symbol (array of integers with the same length as
        'x' or 'y').
    xlim : tuple (floats)
        Plot limits in the X axis.
    ylim : tuple (floats)
        Plot limits in the Y axis.
    xlabel : string
        Character string for label in X axis.
    ylabel : string
        Character string for label in y axis.
    title : string
        Character string for graph title.
    use_r : bool
        If True, the function computes several fits, using R, to
        polynomials of degree deg, deg+1 and deg+2 (when possible).
    show : bool
        If True, the function shows the displayed image. Otherwise
        plt.show() is expected to be executed outside.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.
    pdffile : str or None
        Output PDF file name to save plot. This option is only
        employed when a plot has been generated.
    fig : `~matplotlib.figure.Figure`
        If not None, use this object to generate plot.

    Return
    ------
    poly : instance of Polynomial (numpy)
        Result from the polynomial fit using numpy Polynomial. Only
        points not flagged as rejected are employed in the fit.
    yres : 1d numpy array, float
        Residuals from polynomial fit. Note that the residuals are
        computed for all the points, including the rejected ones. In
        this way the dimension of this array is the same as the
        dimensions of the input 'x' and 'y' arrays.
    reject : 1d numpy array, bool
        Boolean array indicating rejected points.

    """

    # protections
    if type(x) is not np.ndarray:
        raise ValueError("x=" + str(x) + " must be a numpy.ndarray")
    elif x.ndim != 1:
        raise ValueError("x.ndim=" + str(x.ndim) + " must be 1")
    if type(y) is not np.ndarray:
        raise ValueError("y=" + str(y) + " must be a numpy.ndarray")
    elif y.ndim != 1:
        raise ValueError("y.ndim=" + str(y.ndim) + " must be 1")
    npoints = x.size
    if npoints != y.size:
        raise ValueError("x.size != y.size")
    if not isinstance(deg, numbers.Integral):
        raise ValueError("deg=" + str(deg) +
                         " is not a valid integer")
    if deg >= npoints:
        raise ValueError("Polynomial degree=" + str(deg) +
                         " can't be fitted with npoints=" + str(npoints))

    # initialize boolean rejection array
    reject = np.zeros(npoints, dtype=bool)

    # if there is no room to remove any point, compute a fit without
    # rejection
    if deg == npoints - 1:
        poly, yres = polfit_residuals(x=x, y=y, deg=deg, reject=None,
                                      color=color, size=size,
                                      xlim=xlim, ylim=ylim,
                                      xlabel=xlabel, ylabel=ylabel,
                                      title=title,
                                      use_r=use_r,
                                      geometry=geometry,
                                      debugplot=debugplot,
                                      pdffile=pdffile,
                                      fig=fig)
        return poly, yres, reject

    # main loop to reject points iteratively
    loop_to_reject_points = True
    poly = None
    yres = None
    while loop_to_reject_points:
        if abs(debugplot) in [21, 22]:
            poly, yres = polfit_residuals(x=x, y=y, deg=deg, reject=reject,
                                          color=color, size=size,
                                          xlim=xlim, ylim=ylim,
                                          xlabel=xlabel, ylabel=ylabel,
                                          title=title,
                                          use_r=use_r,
                                          show=show,
                                          geometry=geometry,
                                          debugplot=debugplot,
                                          fig=fig)
        else:
            poly, yres = polfit_residuals(x=x, y=y, deg=deg, reject=reject)
        # check that there is room to remove a point with the current
        # polynomial degree
        npoints_effective = npoints - np.sum(reject)
        if deg < npoints_effective - 1:
            # determine robuts standard deviation, excluding points
            # already rejected
            # --- method 1 ---
            # yres_fitted = yres[np.logical_not(reject)]
            # q25, q75 = np.percentile(yres_fitted, q=[25.0, 75.0])
            # rms = 0.7413 * (q75 - q25)
            # --- method 2 ---
            yres_fitted = np.abs(yres[np.logical_not(reject)])
            rms = np.median(yres_fitted)
            if abs(debugplot) >= 10:
                print("--> robust rms:", rms)
            # reject fitted point exceeding the threshold with the
            # largest deviation (note: with this method only one point
            # is removed in each iteration of the loop; this allows the
            # recomputation of the polynomial fit which, sometimes,
            # transforms deviant points into good ones)
            index_to_remove = []
            for i in range(npoints):
                if not reject[i]:
                    if np.abs(yres[i]) > times_sigma_reject * rms:
                        index_to_remove.append(i)
                        if abs(debugplot) >= 10:
                            print('--> suspicious point #', i + 1)
            if len(index_to_remove) == 0:
                if abs(debugplot) >= 10:
                    print('==> no need to remove any point')
                loop_to_reject_points = False
            else:
                imax = np.argmax(np.abs(yres[index_to_remove]))
                reject[index_to_remove[imax]] = True
                if abs(debugplot) >= 10:
                    print('==> removing point #', index_to_remove[imax] + 1)
        else:
            loop_to_reject_points = False

    # plot final fit in case it has not been already shown
    if abs(debugplot) % 10 != 0:
        if abs(debugplot) not in [21, 22]:
            poly, yres = polfit_residuals(x=x, y=y, deg=deg, reject=reject,
                                          color=color, size=size,
                                          xlim=xlim, ylim=ylim,
                                          xlabel=xlabel, ylabel=ylabel,
                                          title=title,
                                          use_r=use_r,
                                          show=show,
                                          geometry=geometry,
                                          debugplot=debugplot,
                                          pdffile=pdffile,
                                          fig=fig)
        else:
            if abs(debugplot) >= 10:
                print(' ')

    # return result
    return poly, yres, reject


def polfit_residuals_with_cook_rejection(
        x, y, deg, times_sigma_cook,
        color='b', size=75,
        xlim=None, ylim=None,
        xlabel=None, ylabel=None, title=None,
        use_r=None,
        show=True,
        geometry=(0, 0, 640, 480),
        debugplot=0,
        pdffile=None,
        fig=None
):
    """Polynomial fit with iterative rejection of points.

    This function makes use of function polfit_residuals for display
    purposes.

    Parameters
    ----------
    x : 1d numpy array, float
        X coordinates of the data being fitted.
    y : 1d numpy array, float
        Y coordinates of the data being fitted.
    deg : int
        Degree of the fitting polynomial.
    times_sigma_cook : float or None
        Number of times the standard deviation of Cook's distances
        above the median value to reject points iteratively.
    color : single character or 1d numpy array of characters
        Color for all the symbols (single character) or for each
        individual symbol (array of color names with the same length as
        'x' or 'y'). If 'color' is a single character, the rejected
        points are displayed in red color, whereas when 'color' is an
        array of color names, rejected points are displayed with the
        color provided in this array.
    size : int
        Marker size for all the symbols (single character) or for each
        individual symbol (array of integers with the same length as
        'x' or 'y').
    xlim : tuple (floats)
        Plot limits in the X axis.
    ylim : tuple (floats)
        Plot limits in the Y axis.
    xlabel : string
        Character string for label in X axis.
    ylabel : string
        Character string for label in y axis.
    title : string
        Character string for graph title.
    use_r : bool
        If True, the function computes several fits, using R, to
        polynomials of degree deg, deg+1 and deg+2 (when possible).
    show : bool
        If True, the function shows the displayed image. Otherwise
        plt.show() is expected to be executed outside.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.
    pdffile : str or None
        Output PDF file name to save plot. This option is only
        employed when a plot has been generated.
    fig : `~matplotlib.figure.Figure`
        If not None, use this object to generate plot.

    Return
    ------
    poly : instance of Polynomial (numpy)
        Result from the polynomial fit using numpy Polynomial. Only
        points not flagged as rejected are employed in the fit.
    yres : 1d numpy array, float
        Residuals from polynomial fit. Note that the residuals are
        computed for all the points, including the rejected ones. In
        this way the dimension of this array is the same as the
        dimensions of the input 'x' and 'y' arrays.
    reject : 1d numpy array, bool
        Boolean array indicating rejected points.

    """

    # protections
    if type(x) is not np.ndarray:
        raise ValueError("x=" + str(x) + " must be a numpy.ndarray")
    elif x.ndim != 1:
        raise ValueError("x.ndim=" + str(x.ndim) + " must be 1")
    if type(y) is not np.ndarray:
        raise ValueError("y=" + str(y) + " must be a numpy.ndarray")
    elif y.ndim != 1:
        raise ValueError("y.ndim=" + str(y.ndim) + " must be 1")
    npoints = x.size
    if npoints != y.size:
        raise ValueError("x.size != y.size")
    if not isinstance(deg, numbers.Integral):
        raise ValueError("deg=" + str(deg) +
                         " is not a valid integer")
    if deg >= npoints:
        raise ValueError("Polynomial degree=" + str(deg) +
                         " can't be fitted with npoints=" + str(npoints))

    # initialize boolean rejection array
    reject = np.zeros(npoints, dtype=bool)

    # if there is no room to remove two points, compute a fit without
    # rejection
    if deg == npoints - 1 or deg == npoints - 2:
        poly, yres = polfit_residuals(x=x, y=y, deg=deg, reject=None,
                                      color=color, size=size,
                                      xlim=xlim, ylim=ylim,
                                      xlabel=xlabel, ylabel=ylabel,
                                      title=title,
                                      use_r=use_r,
                                      show=show,
                                      geometry=geometry,
                                      debugplot=debugplot,
                                      pdffile=pdffile,
                                      fig=fig)
        return poly, yres, reject

    # main loop to reject points iteratively
    loop_to_reject_points = True
    poly = None
    yres = None
    any_point_removed = False
    while loop_to_reject_points:
        # fit to compute residual variance (neglecting already
        # rejected points)
        poly, yres = polfit_residuals(x=x, y=y, deg=deg, reject=reject,
                                      color=color, size=size,
                                      xlim=xlim, ylim=ylim,
                                      xlabel=xlabel, ylabel=ylabel,
                                      title=title,
                                      use_r=use_r,
                                      show=show,
                                      geometry=geometry,
                                      debugplot=debugplot,
                                      pdffile=pdffile,
                                      fig=fig)
        npoints_effective = npoints - np.sum(reject)
        residual_variance = np.sum(yres*yres)/float(npoints_effective-deg-1)
        # check that there is room to remove two points with the
        # current polynomial degree
        if deg <= npoints_effective - 2:
            cook_distance = np.zeros(npoints)
            for i in range(npoints):
                if not reject[i]:
                    reject_cook = np.copy(reject)
                    reject_cook[i] = True
                    poly_cook, yres_cook = polfit_residuals(
                        x=x, y=y, deg=deg, reject=reject_cook,
                        color=color, size=size,
                        xlim=xlim, ylim=ylim,
                        xlabel=xlabel, ylabel=ylabel,
                        title="Computing Cook's distance for point " +
                              str(i+1),
                        use_r=False,
                        debugplot=0)
                    yres_cook_fitted = yres_cook[np.logical_not(reject)]
                    cook_distance[i] = \
                        np.sum(yres_cook_fitted*yres_cook_fitted) / \
                        (2*residual_variance)
                else:
                    cook_distance[i] = np.inf
                if abs(debugplot) >= 10:
                    print('i, cook_distance[i]:', i, cook_distance[i])
            # determine median absolute cook distance, excluding points
            # already rejected
            dist_cook_fitted = np.abs(cook_distance[np.logical_not(reject)])
            q50 = np.median(dist_cook_fitted)
            # rms computed from the previous data after removing the
            # point with the largest deviation
            rms = np.std(np.sort(dist_cook_fitted)[:-2])
            if abs(debugplot) >= 10:
                print("--> median.......:", q50)
                print("--> rms -2 points:", rms)
            # reject fitted point exceeding the threshold with the
            # largest Cook distance (note: with this method only one
            # point is removed in each iteration of the loop). If the
            # polynomial degree is larger than 1, only intermediate
            # points can be discarded (i.e., the first and last point
            # are never rejected because the curvature of the
            # extrapolated polynomials leads to false outliers)
            index_to_remove = []
            if deg > 1:
                n1 = 1
                n2 = npoints - 1
            else:
                n1 = 0
                n2 = npoints
            for i in range(n1, n2):
                if not reject[i]:
                    if np.abs(cook_distance[i]-q50) > times_sigma_cook * rms:
                        index_to_remove.append(i)
                        if abs(debugplot) >= 10:
                            print('--> suspicious point #', i + 1)
            if len(index_to_remove) == 0:
                if abs(debugplot) >= 10:
                    if any_point_removed:
                        print('==> no need to remove any additional point')
                    else:
                        print('==> no need to remove any point')
                loop_to_reject_points = False
            else:
                imax = np.argmax(np.abs(cook_distance[index_to_remove]))
                reject[index_to_remove[imax]] = True
                any_point_removed = True
                if abs(debugplot) >= 10:
                    print('==> removing point #', index_to_remove[imax] + 1)
        else:
            loop_to_reject_points = False

    # return result
    return poly, yres, reject


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(prog='polfit_residuals')
    parser.add_argument("filename",
                        help="ASCII file with data in columns")
    parser.add_argument("col1",
                        help="Column number for X data",
                        type=int)
    parser.add_argument("col2",
                        help="Column number for Y data",
                        type=int)
    parser.add_argument("polydeg",
                        help="Polynomial degree",
                        type=int)
    parser.add_argument("--times_sigma_reject",
                        help="Times sigma to reject points" +
                             " (default=None)",
                        default=None, type=float)
    parser.add_argument("--times_sigma_cook",
                        help="Times sigma to reject points" +
                             " (default=None)",
                        default=None, type=float)
    parser.add_argument("--geometry",
                        help="tuple x,y,dx,dy",
                        default="0,0,640,480")
    parser.add_argument("--pdffile",
                        help="Output PDF file name (default=None)",
                        default=None, type=str)
    args = parser.parse_args(args)

    # ASCII file
    filename = args.filename

    # columns to be plotted (first column will be number 1 and not 0)
    col1 = args.col1 - 1
    col2 = args.col2 - 1

    # polynomial degree
    polydeg = args.polydeg

    # times_sigma_reject
    times_sigma_reject = args.times_sigma_reject

    # times_sigma_cook
    times_sigma_cook = args.times_sigma_cook

    # output PDF file name
    pdffile = args.pdffile

    # geometry
    if args.geometry is None:
        geometry = None
    else:
        tmp_str = args.geometry.split(",")
        x_geom = int(tmp_str[0])
        y_geom = int(tmp_str[1])
        dx_geom = int(tmp_str[2])
        dy_geom = int(tmp_str[3])
        geometry = x_geom, y_geom, dx_geom, dy_geom

    # check
    if times_sigma_reject is not None and times_sigma_cook is not None:
        raise ValueError("ERROR: times_sigma_reject and times_sigma_cook" +
                         " cannot be employed simultaneously")

    # read ASCII file
    bigtable = np.genfromtxt(filename)
    x = bigtable[:, col1]
    y = bigtable[:, col2]

    # plot fit and residuals
    if times_sigma_reject is None and times_sigma_cook is None:
        polfit_residuals(x, y, polydeg, geometry=geometry,
                         debugplot=12, pdffile=pdffile)
    elif times_sigma_reject is not None:
        polfit_residuals_with_sigma_rejection(
            x, y, polydeg, times_sigma_reject,
            geometry=geometry, debugplot=12, pdffile=pdffile
        )
    elif times_sigma_cook is not None:
        polfit_residuals_with_cook_rejection(
            x, y, polydeg, times_sigma_cook,
            geometry=geometry, debugplot=12, pdffile=pdffile
        )


if __name__ == '__main__':
    main()



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

from libc.stdlib cimport malloc, free

from cpython.pycapsule cimport PyCapsule_New
from cpython.pycapsule cimport PyCapsule_GetContext
from cpython.pycapsule cimport PyCapsule_SetContext


# Kernel function is the same
cdef int _kernel_function(double* buffer, int filter_size,
                          double* return_value, void* cb):
    cdef double* data = <double*> cb
    cdef int nmed = filter_size // 2
    cdef int i = 0
    cdef int start = 0
    cdef int mcount = 0
    cdef double th = data[0]
    cdef int limit = <int>data[1]

    if buffer[nmed] < th:
        return_value[0] = 0.0
        return 1

    # Count contiguous equal values to the right
    for i in range(nmed, filter_size-1):
        # print '0-',i, i+1
        # print '0-',buffer[i], buffer[i+1]
        if buffer[i] == buffer[i+1]:
            mcount += 1
            start = i + 1
        else:
            start = i
            break

    for i in range(start, filter_size-1):
        if buffer[i] <= buffer[i+1]:
            return_value[0] = 0.0
            return 1

    # Count contiguous equal values to the left
    for i in range(nmed, 0, -1):
        if buffer[i] == buffer[i-1]:
            mcount += 1
            start = i - 1
        else:
            start = i
            break

    for i in range(0, start):
        if buffer[i] >= buffer[i+1]:
            return_value[0] = 0.0
            return 1

    # Reject peak if it has too much
    # flat values
    if mcount > limit:
        return_value[0] = 0.0
        return 1

    return_value[0] = 1.0
    return 1


cdef void _destructor_cobj(void* cobject, void *kernel_data):
    free(kernel_data)


cdef void _destructor_cap(object cap) noexcept:
    cdef void *cdata
    cdata = PyCapsule_GetContext(cap)
    free(cdata)


def kernel_peak_function(double threshold=0.0, int fpeak=1):

    cdef object result
    cdef double *data

    data = <double*>malloc(2 * sizeof(double))
    if data is NULL:
        raise MemoryError()

    data[0] = threshold
    # A value of 1 allows a peak with 2 equal pixels
    data[1] = fpeak

    result = PyCapsule_New(&_kernel_function,
                           NULL, # if we set a name here, generic_f doesn't work
                           _destructor_cap)

    PyCapsule_SetContext(result, data)

    return result



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import argparse
import numpy as np

from .matplotlib_qt import set_window_geometry
from .pause_debugplot import pause_debugplot


def ximplotxy_jupyter(x, y, fmt=None, **args):
    """Auxiliary function to call ximplotxy from a jupyter notebook.
    """
    using_jupyter = True
    if fmt is None:
        return ximplotxy(x, y, using_jupyter=using_jupyter, **args)
    else:
        return ximplotxy(x, y, fmt, using_jupyter=using_jupyter, **args)


def ximplotxy(x, y, fmt=None, plottype=None,
              xlim=None, ylim=None,
              xlabel=None, ylabel=None, title=None,
              show=True, geometry=(0, 0, 640, 480), tight_layout=True,
              debugplot=0, using_jupyter=False,
              **kwargs):
    """
    Parameters
    ----------
    x : 1d numpy array, float
        Array containing the X coordinate.
    y : 1d numpy array, float
        Array containing the Y coordinate.
    fmt : str, optional
        Format string for quickly setting basic line properties.
    plottype : string
        Plot type. It can be 'loglog', 'semilogx', 'semilogy' or None
        (default, non-logarithmic plot).
    xlim : tuple of floats
        Tuple defining the x-axis range.
    ylim : tuple of floats
        Tuple defining the y-axis range.
    xlabel : string
        X-axis label.
    ylabel : string
        Y-axis label.
    title : string
        Plot title.
    show : bool
        If True, the function shows the displayed image. Otherwise
        plt.show() is expected to be executed outside.
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the window geometry.
    tight_layout : bool
        If True, and show=True, a tight display layout is set.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.
    using_jupyter : bool
        If True, this function is called from a jupyter notebook.

    Returns
    -------
    ax : axes object
        Matplotlib axes instance. This value is returned only when
        'show' is False.

    """

    from numina.array.display.matplotlib_qt import plt
    if not show and using_jupyter:
        plt.ioff()

    fig = plt.figure()
    ax = fig.add_subplot(111)
    if plottype == 'loglog':
        if fmt is None:
            ax.loglog(x, y, **kwargs)
        else:
            ax.loglog(x, y, fmt, **kwargs)
    elif plottype == 'semilogx':
        if fmt is None:
            ax.semilogx(x, y, **kwargs)
        else:
            ax.semilogx(x, y, fmt, **kwargs)
    elif plottype == 'semilogy':
        if fmt is None:
            ax.semilogy(x, y, **kwargs)
        else:
            ax.semilogy(x, y, fmt, **kwargs)
    elif plottype is None:
        if fmt is None:
            ax.plot(x, y, **kwargs)
        else:
            ax.plot(x, y, fmt, **kwargs)
    else:
        raise ValueError('Invalid plottype: ' + str(plottype))

    if xlim is not None:
        ax.set_xlim(xlim[0], xlim[1])
    if ylim is not None:
        ax.set_ylim(ylim[0], ylim[1])

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    set_window_geometry(geometry)

    if show:
        pause_debugplot(debugplot, pltshow=show, tight_layout=tight_layout)
    else:
        if tight_layout:
            plt.tight_layout()
        # return axes
        if using_jupyter:
            plt.ion()
        return ax


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(prog='ximplotxy')
    parser.add_argument("filename",
                        help="ASCII file with data in columns")
    parser.add_argument("col1",
                        help="Column number for X data",
                        type=int)
    parser.add_argument("col2",
                        help="Column number for Y data",
                        type=int)
    parser.add_argument("--kwargs",
                        help="Extra arguments for plot, e.g.: "
                             "\"{'marker':'o',"
                             " 'linestyle':'dotted',"
                             " 'xlabel':'x axis', 'ylabel':'y axis',"
                             " 'title':'sample plot',"
                             " 'xlim':[-1,1], 'ylim':[-2,2],"
                             " 'label':'sample data',"
                             " 'color':'magenta'}\"")
    args = parser.parse_args(args)

    # ASCII file
    filename = args.filename

    # columns to be plotted (first column will be number 1 and not 0)
    col1 = args.col1 - 1
    col2 = args.col2 - 1

    # read ASCII file
    bigtable = np.genfromtxt(filename)
    x = bigtable[:, col1]
    y = bigtable[:, col2]

    if args.kwargs is None:
        ximplotxy(x, y, debugplot=12, marker='o', linestyle='')
    else:
        ximplotxy(x, y, debugplot=12, **eval(args.kwargs))


if __name__ == '__main__':
    main()



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Resampling functions related to wavelength calibration"""

import numpy as np
from numpy.polynomial.polynomial import polyval
from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximplotxy import ximplotxy
from numina.array.interpolation import SteffenInterpolator


def oversample1d(sp, crval1, cdelt1, oversampling=1, debugplot=0):
    """Oversample spectrum.

    Parameters
    ----------
    sp : numpy array
        Spectrum to be oversampled.
    crval1 : float
        Abscissae of the center of the first pixel in the original
        spectrum 'sp'.
    cdelt1 : float
        Abscissae increment corresponding to 1 pixel in the original
        spectrum 'sp'.
    oversampling : int
        Oversampling value per pixel.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    sp_over : numpy array
        Oversampled data array.
    crval1_over : float
        Abscissae of the center of the first pixel in the oversampled
        spectrum.
    cdelt1_over : float
        Abscissae of the center of the last pixel in the oversampled
        spectrum.

    """

    if sp.ndim != 1:
        raise ValueError('Unexpected array dimensions')

    naxis1 = sp.size
    naxis1_over = naxis1 * oversampling
    cdelt1_over = cdelt1 / oversampling
    xmin = crval1 - cdelt1/2   # left border of first pixel
    crval1_over = xmin + cdelt1_over / 2

    sp_over = np.zeros(naxis1_over)
    for i in range(naxis1):
        i1 = i * oversampling
        i2 = i1 + oversampling
        sp_over[i1:i2] = sp[i]

    if abs(debugplot) in (21, 22):
        crvaln = crval1 + (naxis1 - 1) * cdelt1
        crvaln_over = crval1_over + (naxis1_over - 1) * cdelt1_over
        xover = np.linspace(crval1_over, crvaln_over, naxis1_over)
        ax = ximplotxy(np.linspace(crval1, crvaln, naxis1), sp, 'bo',
                       label='original', show=False)
        ax.plot(xover, sp_over, 'r+', label='resampled')
        pause_debugplot(debugplot, pltshow=True)

    return sp_over, crval1_over, cdelt1_over


def rebin(a, *args):
    """See http://scipy-cookbook.readthedocs.io/items/Rebinning.html

    Note: integer division in the computation of 'factor' has been
    included to avoid the following runtime message:
    VisibleDeprecationWarning: using a non-integer number instead of
    an integer will result in an error in the future


    """

    shape = a.shape
    len_shape = len(shape)
    factor = np.asarray(shape) // np.asarray(args)  # noqa: F841
    # ev_list = ['a.reshape('] + \
    #           ['args[%d], factor[%d], ' % (i, i) for i in range(len_shape)] + \
    #           [')'] + ['.mean(%d)' % (i+1) for i in range(len_shape)]
    # FIXME: this construction is weird
    ev_list = ['a.reshape('] + \
              [f'args[{idx}], factor[{idx}], ' for idx in range(len_shape)] + \
              [')'] + [f'.mean({idx + 1})' for idx in range(len_shape)]
    # print(''.join(ev_list))
    return eval(''.join(ev_list))


def shiftx_image2d_flux(image2d_orig, xoffset):
    """Resample 2D image using a shift in the x direction (flux is preserved).

    Parameters
    ----------
    image2d_orig : numpy array
        2D image to be resampled.
    xoffset : float
        Offset to be applied.

    Returns
    -------
    image2d_resampled : numpy array
        Resampled 2D image.

    """

    if image2d_orig.ndim == 1:
        naxis1 = image2d_orig.size
    elif image2d_orig.ndim == 2:
        naxis2, naxis1 = image2d_orig.shape
    else:
        print('>>> image2d_orig.shape:', image2d_orig.shape)
        raise ValueError('Unexpected number of dimensions')

    return resample_image2d_flux(image2d_orig,
                                 naxis1=naxis1,
                                 cdelt1=1,
                                 crval1=1,
                                 crpix1=1,
                                 coeff=[xoffset, 1])


def resample_image2d_flux(image2d_orig,
                          naxis1, cdelt1, crval1, crpix1, coeff):
    """Resample a 1D/2D image using NAXIS1, CDELT1, CRVAL1, and CRPIX1.

    The same NAXIS1, CDELT1, CRVAL1, and CRPIX1 are employed for all
    the scans (rows) of the original 'image2d'. The wavelength
    calibrated output image has dimensions NAXIS1 * NSCAN, where NSCAN
    is the original number of scans (rows) of the original image.

    Flux is preserved.

    Parameters
    ----------
    image2d_orig : numpy array
        1D or 2D image to be resampled.
    naxis1 : int
        NAXIS1 of the resampled image.
    cdelt1 : float
        CDELT1 of the resampled image.
    crval1 : float
        CRVAL1 of the resampled image.
    crpix1 : float
        CRPIX1 of the resampled image.
    coeff : numpy array
        Coefficients of the wavelength calibration polynomial.

    Returns
    -------
    image2d_resampled : numpy array
        Wavelength calibrated 1D or 2D image.

    """

    # duplicate input array, avoiding problems when using as input
    # 1d numpy arrays with shape (nchan,) instead of a 2d numpy
    # array with shape (1,nchan)
    if image2d_orig.ndim == 1:
        nscan = 1
        nchan = image2d_orig.size
        image2d = np.zeros((nscan, nchan))
        image2d[0, :] = np.copy(image2d_orig)
    elif image2d_orig.ndim == 2:
        nscan, nchan = image2d_orig.shape
        image2d = np.copy(image2d_orig)
    else:
        print('>>> image2d_orig.shape:', image2d_orig.shape)
        raise ValueError('Unexpected number of dimensions')

    new_x = np.arange(naxis1)
    new_wl = crval1 + cdelt1 * new_x

    old_x_borders = np.arange(-0.5, nchan)
    old_x_borders += crpix1  # following FITS criterium

    new_borders = map_borders(new_wl)

    accum_flux = np.empty((nscan, nchan + 1))
    accum_flux[:, 1:] = np.cumsum(image2d, axis=1)
    accum_flux[:, 0] = 0.0
    image2d_resampled = np.zeros((nscan, naxis1))

    old_wl_borders = polyval(old_x_borders, coeff)

    for iscan in range(nscan):
        # We need a monotonic interpolator
        # linear would work, we use a cubic interpolator
        interpolator = SteffenInterpolator(
            old_wl_borders,
            accum_flux[iscan],
            extrapolate='border'
        )
        fl_borders = interpolator(new_borders)
        image2d_resampled[iscan] = fl_borders[1:] - fl_borders[:-1]

    if image2d_orig.ndim == 1:
        return image2d_resampled[0, :]
    else:
        return image2d_resampled


def map_borders(wls):
    """Compute borders of pixels for interpolation.

    The border of the pixel is assumed to be midway of the wls
    """
    midpt_wl = 0.5 * (wls[1:] + wls[:-1])
    all_borders = np.zeros((wls.shape[0] + 1,))
    all_borders[1:-1] = midpt_wl
    all_borders[0] = 2 * wls[0] - midpt_wl[0]
    all_borders[-1] = 2 * wls[-1] - midpt_wl[-1]
    return all_borders



#
# Copyright 2014-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import uuid

from numina.types.qc import QC

from .unserial import unserial


class StoredProduct(object):
    """A product returned from the DAL"""

    def __init__(self, id, content, tags, **kwds):
        self.id = id
        self.content = content
        self.tags = tags


class StoredParameter(object):
    """A parameter returned from the DAL"""

    def __init__(self, content):
        self.content = content


class StoredResult(object):
    """Recover the RecipeResult values stored in the Backend"""

    def __init__(self):
        self.qc = QC.UNKNOWN
        self.uuid = uuid.UUID('00000000-0000-0000-0000-000000000000')

    @classmethod
    def load_data(cls, state):
        obj = cls.__new__(cls)
        obj._from_dict(state)
        return obj

    def _from_dict(self, state):
        self.qc = QC[state.get('qc', 'UNKNOWN')]
        if 'uuid' in state:
            self.uuid = uuid.UUID(state['uuid'])

        values = state.get('values', {})
        if isinstance(values, list):
            values = {o['name']: o['content'] for o in values}
        for key, val in values.items():
            loaded = unserial(val)
            setattr(self, key, loaded)


# A translation of the entries of oblocks
# Notice that this is different to ObservationResult
# that contains the results of the reductions
class ObservingBlock(object):
    def __init__(self, id, instrument, mode, images, children, parent=None, facts=None):
        self.id = id
        self.instrument = instrument
        self.mode = mode
        # only one of files and children can
        # be different from []
        self.images = images
        self.frames = self.images
        self.children = children
        self.parent = parent
        self.facts = facts



#
# Copyright 2008-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Different methods for combining lists of arrays."""

import numpy

import numina.array._combine as intl  # noqa
from numina.array.crmasks.apply_crmasks import apply_crmasks


def mean(arrays, masks=None, dtype=None, out=None, out_res=None,
         out_var=None, out_pix=None, zeros=None, scales=None, weights=None):
    """Combine arrays using the mean, with masks and offsets.

    Arrays and masks are a list of array objects. All input arrays
    have the same shape. If present, the masks have the same shape
    also.

    The function returns an array with one more dimension than the
    inputs and with size (3, shape). out[0] contains the mean,
    out[1] the variance and out[2] the number of points used.

    :param arrays: a list of arrays
    :param masks: a list of mask arrays, True values are masked
    :param dtype: data type of the output
    :param out: optional output, with one more axis than the input arrays
    :return: mean, variance of the mean and number of points stored

    Example:
        >>> import numpy
        >>> image = numpy.array([[1., 3.], [1., -1.4]])
        >>> inputs = [image, image + 1]
        >>> mean(inputs)
        array([[[ 1.5,  3.5],
                [ 1.5, -0.9]],
        <BLANKLINE>
               [[ 0.5,  0.5],
                [ 0.5,  0.5]],
        <BLANKLINE>
               [[ 2. ,  2. ],
                [ 2. ,  2. ]]])

    """
    return generic_combine(intl.mean_method(), arrays, masks=masks, dtype=dtype, out=out, out_res=out_res,
                           out_var=out_var, out_pix=out_pix, zeros=zeros, scales=scales, weights=weights)


def median(arrays, masks=None, dtype=None, out=None, out_res=None,
           out_var=None, out_pix=None, zeros=None, scales=None, weights=None):
    """Combine arrays using the median, with masks.

    Arrays and masks are a list of array objects. All input arrays
    have the same shape. If present, the masks have the same shape
    also.

    The function returns an array with one more dimension than the
    inputs and with size (3, shape). out[0] contains the mean,
    out[1] the variance and out[2] the number of points used.

    :param arrays: a list of arrays
    :param masks: a list of mask arrays, True values are masked
    :param dtype: data type of the output
    :param out: optional output, with one more axis than the input arrays

    :return: median, variance of the median and number of points stored
    """

    return generic_combine(intl.median_method(), arrays, masks=masks, dtype=dtype, out=out,
                           out_res=out_res, out_var=out_var, out_pix=out_pix,
                           zeros=zeros, scales=scales, weights=weights)


def sigmaclip(arrays, masks=None, dtype=None, out=None, out_res=None, out_var=None,
              out_pix=None, zeros=None, scales=None, weights=None, low=3.0, high=3.0):
    """Combine arrays using the sigma-clipping, with masks.

    Inputs and masks are a list of array objects. All input arrays
    have the same shape. If present, the masks have the same shape
    also.

    The function returns an array with one more dimension than the
    inputs and with size (3, shape). out[0] contains the mean,
    out[1] the variance and out[2] the number of points used.

    :param arrays: a list of arrays
    :param masks: a list of mask arrays, True values are masked
    :param dtype: data type of the output
    :param out: optional output, with one more axis than the input arrays
    :param low:
    :param high:
    :return: mean, variance of the mean and number of points stored
    """

    return generic_combine(intl.sigmaclip_method(low, high), arrays, out=out, dtype=dtype, out_res=out_res,
                           out_var=out_var, out_pix=out_pix, masks=masks, zeros=zeros, scales=scales, weights=weights)


def minmax(arrays, masks=None, dtype=None, zeros=None, scales=None, weights=None, nmin=1, nmax=1):
    """Combine arrays using mix max rejection, with masks.

    Inputs and masks are a list of array objects. All input arrays
    have the same shape. If present, the masks have the same shape
    also.

    The function returns an array with one more dimension than the
    inputs and with size (3, shape). out[0] contains the mean,
    out[1] the variance and out[2] the number of points used.

    :param arrays: a list of arrays
    :param masks: a list of mask arrays, True values are masked
    :param dtype: data type of the output
    :param out: optional output, with one more axis than the input arrays
    :param nmin:
    :param nmax:
    :return: mean, variance of the mean and number of points stored
    """

    return generic_combine(intl.minmax_method(nmin, nmax), arrays, dtype=dtype,
                           zeros=zeros, scales=scales, weights=weights)


def quantileclip(arrays, masks=None, out_res=None, out_var=None, out_pix=None, out=None,
                 dtype=None, zeros=None, scales=None, weights=None, fclip=0.10):
    """Combine arrays using the sigma-clipping, with masks.

    Inputs and masks are a list of array objects. All input arrays
    have the same shape. If present, the masks have the same shape
    also.

    The function returns an array with one more dimension than the
    inputs and with size (3, shape). out[0] contains the mean,
    out[1] the variance and out[2] the number of points used.

    :param arrays: a list of arrays
    :param masks: a list of mask arrays, True values are masked
    :param out: optional output, with one more axis than the input arrays
    :param fclip: fraction of points removed on both ends. Maximum is 0.4 (80% of points rejected)
    :return: mean, variance of the mean and number of points stored
    """
    return generic_combine(intl.quantileclip_method(fclip), arrays, masks=masks, out=out, dtype=dtype,
                           out_res=out_res, out_var=out_var, out_pix=out_pix,
                           zeros=zeros, scales=scales, weights=weights)


def flatcombine(arrays, masks=None, scales=None, dtype=None, low=3.0, high=3.0, blank=1.0):
    """Combine flat arrays.

    :param arrays: a list of arrays
    :param masks: a list of mask arrays, True values are masked
    :param dtype: data type of the output
    :param out: optional output, with one more axis than the input arrays
    :param blank: non-positive values are substituted by this on output
    :return: mean, variance of the mean and number of points stored
    """

    result = sigmaclip(arrays, masks=masks, dtype=dtype, scales=scales, low=low, high=high)

    # Substitute values <= 0 by blank
    mm = result[0] <= 0
    result[0, mm] = blank
    # Add values to mask
    result[1:2, mm] = 0

    return result


def zerocombine(arrays, masks, scales=None, dtype=None):
    """Combine zero arrays.

    :param arrays: a list of arrays
    :param masks: a list of mask arrays, True values are masked
    :param dtype: data type of the output
    :param scales:
    :return: median, variance of the median and number of points stored
    """

    result = median(arrays, masks=masks, scales=scales, dtype=dtype)

    return result


def sum(arrays, masks=None, out_res=None, out_var=None, out_pix=None, out=None, dtype=None, zeros=None, scales=None):
    """Combine arrays by addition, with masks and offsets.

    Arrays and masks are a list of array objects. All input arrays
    have the same shape. If present, the masks have the same shape
    also.

    The function returns an array with one more dimension than the
    inputs and with size (3, shape). out[0] contains the sum,
    out[1] the variance and out[2] the number of points used.

    :param arrays: a list of arrays
    :param masks: a list of mask arrays, True values are masked
    :param out: optional output, with one more axis than the input arrays
    :return: sum, variance of the sum and number of points stored

    Example:
        >>> import numpy
        >>> image = numpy.array([[1., 3.], [1., -1.4]])
        >>> inputs = [image, image + 1]
        >>> sum(inputs)
        array([[[ 1.5,  3.5],
                [ 1.5, -0.9]],
        <BLANKLINE>
               [[ 0.5,  0.5],
                [ 0.5,  0.5]],
        <BLANKLINE>
               [[ 2. ,  2. ],
                [ 2. ,  2. ]]])

    """
    return generic_combine(intl.sum_method(), arrays, out=out, dtype=dtype, out_res=out_res,
                           out_var=out_var, out_pix=out_pix, masks=masks, zeros=zeros, scales=scales)


def generic_combine(method, arrays, out_res=None, out_var=None, out_pix=None, out=None,
                    dtype=None, masks=None, zeros=None, scales=None, weights=None):

    # Trying to use all the arguments
    if out is not None:
        out_res = out[0]
        out_var = out[1]
        out_pix = out[2]
    else:
        if dtype is not None and len(arrays) > 0:
            out_res = numpy.empty_like(arrays[0], dtype=dtype)
            out_var = numpy.empty_like(arrays[0], dtype=dtype)
            out_pix = numpy.empty_like(arrays[0], dtype=dtype)

    return intl.generic_combine(method, arrays, out_res=out_res, out_var=out_var,
                                out_pix=out_pix, masks=masks, zeros=zeros, scales=scales)


def mediancr(arrays, crmasks, dtype, use_lamedian=False, apply_flux_factor=None, bias=None):
    """Combine arrays using the median with cosmic ray mask.

    The function returns the median of the input arrays, applying
    a cosmic ray mask computed to detect cosmis rays hitting more that
    once in a single pixel. The masked pixels are replaced by the minimum
    value of the same pixel in the input arrays.
    """
    return apply_crmasks(
        list_arrays=arrays,
        hdul_masks=crmasks,
        use_lamedian=use_lamedian,
        combination='mediancr',
        dtype=dtype,
        apply_flux_factor=apply_flux_factor,
        bias=bias
    )


def meancr(arrays, crmasks, dtype, use_lamedian=False, apply_flux_factor=None, bias=None):
    """Combine arrays using the mean with individual cosmic ray masks.

    The function returns the mean of the input arrays, applying
    a cosmic ray mask computed for each input array. This allows to employ
    numpy masked arrays to determine the mean of each pixel. For those
    pixels where the cosmic ray mask is True in all the input arrays,
    the minimum value of the same pixel in the input arrays is used
    as the value of the pixel in the output.
    """
    return apply_crmasks(
        list_arrays=arrays,
        hdul_masks=crmasks,
        use_lamedian=use_lamedian,
        combination='meancr',
        dtype=dtype,
        apply_flux_factor=apply_flux_factor,
        bias=bias
    )


def meancrt(arrays, crmasks, dtype, use_lamedian=False, apply_flux_factor=None, bias=None):
    """Combine arrays using the mean with a single cosmic ray mask.

    The function returns the mean of the input arrays, applying
    a cosmic ray mask computed at once, by stacking the input arrays
    and trying to detect all the cosmic rays in all the exposures piled
    up in a single image. The masked pixels are replaced by the value
    of the same pixel when the combination is computed using 'mediancr'.
    """
    return apply_crmasks(
        list_arrays=arrays,
        hdul_masks=crmasks,
        use_lamedian=use_lamedian,
        combination='meancrt',
        dtype=dtype,
        apply_flux_factor=apply_flux_factor,
        bias=bias
    )



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy as np
from numpy.testing import assert_almost_equal, assert_allclose
from astropy.modeling.models import Gaussian2D

from ...constants import FWHM_G
from ..fwhm import compute_fwhm_2d_simple
from ..fwhm import compute_fwhm_1d_simple


def test_fwhm_2d_simple():
    xcenter = 120.2
    ycenter = 122.3
    xsig = 25.2
    ysig = 12.8
    g2d_model = Gaussian2D(amplitude=1.0, x_mean=xcenter,
                           y_mean=ycenter, x_stddev=xsig, y_stddev=ysig)
    y, x = np.mgrid[:250, :250]
    img = g2d_model(x, y)
    _peak, fwhmx, fwhmy = compute_fwhm_2d_simple(img, xcenter, ycenter)
    assert_allclose(fwhmx, FWHM_G * xsig, rtol=1e-3)
    assert_allclose(fwhmy, FWHM_G * ysig, rtol=1e-3)


def test_fwhm_1d_simple():
    # Test with square box
    ref_peak = 1.0
    ref_val = 6.0

    Fr_ref = np.zeros((15,))
    Fr_ref[4:9] = ref_peak

    peak, fwhm = compute_fwhm_1d_simple(Fr_ref, ref_val)

    assert_almost_equal(peak, ref_peak)
    assert_allclose(fwhm, 5.0)

    # Test with a gaussian
    rad = np.arange(0, 250, 1.0)
    center = 120.2
    sigma = 25.3
    Fr_ref = np.exp(-0.5 * ((rad - center) / sigma)**2)

    peak, fwhm = compute_fwhm_1d_simple(Fr_ref, center)
    assert_allclose(fwhm, FWHM_G * sigma, rtol=1e-4)

    # Test with a gaussian, not starting in 0
    rad = np.arange(10, 260, 1.0)
    center = 130.2
    sigma = 25.3
    Fr_ref = np.exp(-0.5 * ((rad - center) / sigma)**2)

    peak, fwhm = compute_fwhm_1d_simple(Fr_ref, center, rad)
    assert_allclose(fwhm, FWHM_G * sigma, rtol=1e-4)



#
# Copyright 2019-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Description of the origin of different configurations"""

from datetime import datetime
import uuid
import typing

from dateutil.parser import isoparse


class UndefinedOrigin:
    """Origin not defined"""

    def is_valid_date(self, _):
        return True


class ElementOrigin:
    """Description of the origin of a particular configuration"""

    def __init__(
        self,
        name: str,
        uuid_in: uuid.UUID | str,
        date_start: str | datetime | None = None,
        date_end: str | datetime | None = None,
        description: str = "",
    ):
        self.name = name

        if isinstance(uuid_in, str):
            self.uuid = uuid.UUID(uuid_in)

        self.date_start = self._isoparse(date_start)
        self.date_end = self._isoparse(date_end)

        self.description = description

    @staticmethod
    def _isoparse(date: str | datetime):
        if isinstance(date, str):
            return isoparse(date)
        else:
            return date

    def is_valid_date(self, cdate: str | None) -> bool:
        """Check if the element is valid for a given date"""

        if cdate is None:
            return True

        cdate = self._isoparse(cdate)

        if self.date_start:
            after_start = cdate >= self.date_start
        else:
            after_start = True

        if self.date_end:
            before_end = cdate < self.date_end
        else:
            before_end = True

        if after_start and before_end:
            return True
        else:
            return False

    @classmethod
    def from_dict(cls, val: dict[str, typing.Any]) -> "ElementOrigin":
        return cls.from_keys(**val)

    @classmethod
    def from_keys(cls, **kwargs) -> "ElementOrigin":

        # Parse kwargs
        name = kwargs["name"]
        desc = kwargs.get("description", "")
        uuid = kwargs["uuid"]
        if kwargs["date_start"]:
            date_start = isoparse(kwargs["date_start"])
        else:
            date_start = None

        if kwargs["date_end"]:
            date_end = isoparse(kwargs["date_end"])
        else:
            date_end = None
        return ElementOrigin(name, uuid, date_start, date_end, desc)



# Compatibility

import warnings

from numina.types.array import ArrayNType  # noqa: F401
from numina.types.array import ArrayType  # noqa: F401
from numina.types.product import DataProductMixin  # noqa: F401
from numina.types.product import DataProductTag  # noqa: F401
from numina.types.linescatalog import LinesCatalog  # noqa: F401

warnings.warn("deprecated, use numina.types instead",
              DeprecationWarning, stacklevel=2)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import argparse
import glob
import os.path


class FileInfo(object):
    """Auxiliary class to store filename and associated information.

    Parameters
    ----------
    filename : string
        File name.
    fileinfo : list of strings or None
        Associated file information.

    Attributes
    ----------
    The same as the parameters.

    """

    def __init__(self, filename, extnum=None, fileinfo=None):
        self.filename = filename
        self.extnum = extnum
        self.fileinfo = fileinfo

    def __str__(self):
        """Printable representation of a FileInfo instance."""

        output = "<FileInfo instance>\n" + \
                 "- filename: " + self.filename + "\n" + \
                 "- extnum..: " + self.extnum + "\n" + \
                 "- fileinfo: " + str(self.fileinfo)

        return output


def check_extnum(filename):
    """Return extension number when given as filename.fits[extnum]

    Parameters
    ----------
    filename : string
        File name.

    Returns
    -------
    filename : string
        File name without the extension number (if initially present).
    extnum : int or None
        Extension number.

    """

    extnum = None
    if filename[-1] == ']':
        leftbracket = filename.rfind('[')
        if leftbracket != -1:
            cdum = filename[(leftbracket + 1):-1]
            if len(cdum) > 0:
                try:
                    extnum = int(cdum)
                except ValueError:
                    raise ValueError("Invalid extension number {}".format(
                        filename[(leftbracket + 1):-1]
                    ))
                # remove extension number from file name
                filename = filename[:leftbracket]

    return filename, extnum


def list_fileinfo_from_txt(filename):
    """Returns a list of FileInfo instances if filename is a TXT file.

    Parameters
    ----------
    filename : string
        Name of a file (wildcards are acceptable) or a TXT file
        containing a list of files. Empty Lines, and lines starting by
        a hash or a at symbol in the TXT file are ignored.

    Returns
    -------
    output : list of FileInfo instances
        List of FileInfo instances containing the name of the files
        and additional file information.

    """

    # check for input file
    if not os.path.isfile(filename):
        # check for wildcards
        list_fits_files = glob.glob(filename)
        if len(list_fits_files) == 0:
            # check for extension at the end of the file name
            tmpfile, tmpextnum = check_extnum(filename)
            list_fits_files = glob.glob(tmpfile)
            if len(list_fits_files) == 0:
                raise ValueError("File " + filename + " not found!")
            else:
                list_fits_files.sort()
                output = [FileInfo(tmpfile, tmpextnum) for
                          tmpfile in list_fits_files]
                return output
        else:
            list_fits_files.sort()
            output = [FileInfo(tmpfile) for tmpfile in list_fits_files]
            return output

    # if input file is a txt file, assume it is a list of FITS files
    if filename[-4:] == ".txt":
        with open(filename) as f:
            file_content = f.read().splitlines()
        output = []
        for line in file_content:
            if len(line) > 0:
                if line[0] not in ['#', '@']:
                    tmplist = line.split()
                    tmpfile = tmplist[0]
                    tmpextnum = None
                    if len(tmplist) > 1:
                        tmpinfo = tmplist[1:]
                    else:
                        tmpinfo = None
                    if not os.path.isfile(tmpfile):
                        # check for extension
                        tmpfile, tmpextnum = check_extnum(tmpfile)
                        if tmpextnum is None:
                            raise ValueError(f"File {tmpfile} not found")
                        else:
                            if not os.path.isfile(tmpfile):
                                raise ValueError(f"File {tmpfile} not found")

                    output.append(FileInfo(tmpfile, tmpextnum, tmpinfo))
    else:
        output = [FileInfo(filename)]

    return output


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(prog='fileinfo')
    parser.add_argument("txt_file",
                        help="txt file with list files")
    args = parser.parse_args(args)

    # execute function
    list_fileinfo = list_fileinfo_from_txt(args.txt_file)
    for item in list_fileinfo:
        print(item)


if __name__ == "__main__":

    main()



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Compute offsets between 2d images using 2D cross-correlation."""

import numpy as np
from scipy.signal import correlate2d

import numina.array.imsurfit as imsurfit
from numina.array.imsurfit import vertex_of_quadratic
import numina.array.utils as utils


def yx_offsets_correlate2d(reference_image, moving_image, refine_box=3):
    """Compute (Y, X) offsets between two 2D images using cross-correlation.

    Here we use two calls to correlate2d, and this is itentional.
    The first call computes the autocorrelation of the reference image
    with itself. The second call computes the actual cross-correlation
    between the reference image and the moving image. Both correlation
    maps (corr_self and corr) are locally fitted with a 2D quadratic
    surface (via imsurfit) around their respective peaks. The difference
    between the two fitted peaks provides a high-precision (subpixel)
    estimate of the offsets between the two images. This approach helps
    to mitigate potential biases that could arise from asymmetries or
    noise in the images, leading to a more robust determination of the
    offsets.

    Parameters
    ----------
    reference_image : 2D array
        The reference image (e.g., a template or a fixed image).
    moving_image : 2D array
        The image to be aligned (e.g., a target or moving image).
    refine_box : int, optional
        The size of the box (in pixels) around the peak of the correlation
        to use for the quadratic fit. Default is 3.

    Returns
    -------
    yx_offsets : array
        A 2-element array containing the (Y, X) offsets between the two images.

    """

    corr_self = correlate2d(
        in1=reference_image,
        in2=reference_image,
        mode='full',
        boundary='fill',
        fillvalue=0
    )
    corr = correlate2d(
        in1=reference_image,
        in2=moving_image,
        mode='full',
        boundary='fill',
        fillvalue=0
    )
    maxindex_self = np.unravel_index(np.argmax(corr_self), corr_self.shape)
    maxindex = np.unravel_index(np.argmax(corr), corr.shape)
    region_refine_self = utils.image_box(
        maxindex_self,
        corr_self.shape,
        box=(refine_box, refine_box)
    )
    region_refine = utils.image_box(
        maxindex,
        corr.shape,
        box=(refine_box, refine_box)
    )
    coeffs_self, = imsurfit.imsurfit(corr_self[region_refine_self], order=2)
    coeffs, = imsurfit.imsurfit(corr[region_refine], order=2)
    xm, ym = vertex_of_quadratic(coeffs_self)
    maxindex_self += np.asarray([ym, xm])
    xm, ym = vertex_of_quadratic(coeffs)
    maxindex += np.asarray([ym, xm])
    yx_offsets = np.asarray(maxindex) - np.asarray(maxindex_self)

    return yx_offsets



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

def overplot_ds9reg(filename, ax):
    """Overplot a ds9 region file.

    Parameters
    ----------
    filename : str
        File name of the ds9 region file.
    ax : matplotlib axes instance
        Matplotlib axes instance.

    """

    # read ds9 region file
    with open(filename) as f:
        file_content = f.read().splitlines()

    # check first line
    first_line = file_content[0]
    if "# Region file format: DS9" not in first_line:
        raise ValueError("Unrecognized ds9 region file format")

    for line in file_content:
        if line[0:4] == "line":
            line_fields = line.split()
            x1 = float(line_fields[1])
            y1 = float(line_fields[2])
            x2 = float(line_fields[3])
            y2 = float(line_fields[4])
            if "color" in line:
                i = line.find("color=")
                color = line[i+6:i+13]
            else:
                color = "green"
            ax.plot([x1, x2], [y1, y2], '-', color=color)
        elif line[0:4] == "text":
            line_fields = line.split()
            x0 = float(line_fields[1])
            y0 = float(line_fields[2])
            text = line_fields[3][1:-1]
            if "color" in line:
                i = line.find("color=")
                color = line[i+6:i+13]
            else:
                color = "green"
            ax.text(x0, y0, text, fontsize=8,
                    bbox=dict(boxstyle="round,pad=0.1",
                              fc="white", ec="grey", ),
                    color=color, fontweight='bold', backgroundcolor='white',
                    ha='center')
        else:
            # ignore
            pass



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Remove isolated pixels in a 2D array."""
import numpy as np


def remove_isolated_pixels(h):
    """Remove isolated pixels in a 2D array corresponding to a 2D histogram.

    Parameters
    ----------
    h : 2D numpy array
        The input 2D array (e.g., a histogram).

    Returns
    -------
    hclean : 2D numpy array
        The cleaned 2D array with isolated pixels removed.
    """
    if not isinstance(h, np.ndarray) or h.ndim != 2:
        raise ValueError("Input must be a 2D numpy array.")

    naxis2, naxis1 = h.shape
    flag = np.zeros_like(h, dtype=float)
    # Fill isolated holes
    for i in range(naxis2):
        for j in range(naxis1):
            if h[i, j] == 0:
                k, ktot = 0, 0
                fsum = 0.0
                for ii in [i-1, i, i+1]:
                    if 0 <= ii < naxis2:
                        for jj in [j-1, j, j+1]:
                            if 0 <= jj < naxis1:
                                ktot += 1
                                if h[ii, jj] != 0:
                                    k += 1
                                    fsum += h[ii, jj]
                if k == ktot - 1:
                    flag[i, j] = fsum / k
    hclean = h.copy()
    hclean[flag > 0] = flag[flag > 0]

    # Remove pixels with less than 4 neighbors
    flag = np.zeros_like(h, dtype=np.uint8)
    for i in range(naxis2):
        for j in range(naxis1):
            if h[i, j] != 0:
                k = 0
                for ii in [i-1, i, i+1]:
                    if 0 <= ii < naxis2:
                        for jj in [j-1, j, j+1]:
                            if 0 <= jj < naxis1:
                                if h[ii, jj] != 0:
                                    k += 1
                if k < 5:
                    flag[i, j] = 1
    hclean[flag == 1] = 0

    """
    # Remove pixels with no neighbor on the left hand side
    # when moving from left to right in each row
    for i in range(naxis2):
        j = 0
        loop = True
        while loop:
            j += 1
            if j < naxis1:
                loop = h[i, j] != 0
            else:
                loop = False
        if j < naxis1:
            pass
            hclean[i, j:] = 0
    """

    return hclean



def test_box_prop1(pattri_ins):
    assert pattri_ins.get_property("PSU.spaces") == [{"mode_a_par": -20}]


def test_ins_prop1(pattri_ins):
    # assert pattri_ins._internal_state == {'key2': 'A'}
    assert pattri_ins.get_property("PATTRI.key20") == -1.0
    # assert pattri_ins._internal_state == {'key2': 'A'}
    pattri_ins.key2 = "B"
    # assert pattri_ins._internal_state == {'key2': 'B'}
    assert pattri_ins.get_property("PATTRI.key20") == 1


def test_ins_prop2(pattri_ins):
    # assert pattri_ins._internal_state == {'key2': 'A'}
    assert pattri_ins.key3 == "B"
    # assert pattri_ins._internal_state == {'key2': 'A', 'key3': 'B'}


def test_ins_prop3(pattri_ins):
    assert pattri_ins.get_property("PATTRI.key21") == -2.0
    assert pattri_ins.get_device("PSU")._internal_state["tag0"] == "tag0_A"
    pattri_ins.get_device("PSU").tag0 = "tag0_B"
    assert pattri_ins.get_device("PSU")._internal_state["tag0"] == "tag0_B"
    assert pattri_ins.get_property("PATTRI.key21") == 2.0



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""

import numina.drps
from numina.user.clishowins import print_no_instrument


def register(subparsers, config):
    parser_show_mode = subparsers.add_parser(
        'show-modes',
        help='show information of observing modes'
    )

    parser_show_mode.set_defaults(
        command=show_observingmodes,
        verbose=0,
        what='om'
    )
#    parser_show_mode.add_argument('--verbose', '-v', action='count')
    parser_show_mode.add_argument(
        '-i', '--instrument',
        help='filter modes by instrument'
    )

    parser_show_mode.add_argument(
        'name', nargs='*', default=None,
        help='filter observing modes by name'
    )

    return parser_show_mode


def show_observingmodes(args, extra_args, config):

    drpsys = numina.drps.get_system_drps()

    if args.instrument:
        name = args.instrument
        try:
            val = drpsys.query_by_name(name)
        except KeyError:
            val = None
        res = [(name, val)]
    else:
        res = drpsys.query_all().items()

    for name, theins in res:
        if theins:
            for mode in theins.modes.values():
                if not args.name or (mode.key in args.name):
                    print_obsmode(mode, theins)
        else:
            print_no_instrument(name)


def print_obsmode(obsmode, instrument, ins=False):
    print(f'Observing Mode: {obsmode.name!r} ({obsmode.key})')
    print(' summary:', obsmode.summary)
    print(' instrument:', instrument.name)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""A class to mock the DRP loading process."""

import importlib.metadata
import sys

import backports.entry_points_selectable
import numina.core.pipelineload as pload


def create_mock_entry_point(monkeypatch, entry_name, drploader):

    value = f"{entry_name}.loader"
    group = "numina.pipeline.1"

    # In python >= 3.11 EntryPoint is inmutable
    # and we cannot use monkeypatch
    # We remove __setattr__ that raises an exception
    if not sys.version_info < (3, 11):
        try:
            delattr(importlib.metadata.EntryPoint, '__setattr__')
        except AttributeError:
            pass

    ep = importlib.metadata.EntryPoint(
        name=entry_name, value=value, group=group)

    monkeypatch.setattr(ep, 'load', lambda: drploader)

    return ep


class DRPMocker(object):
    """Mocks the DRP loading process for testing."""

    def __init__(self, monkeypatch):
        self.monkeypatch = monkeypatch
        self._eps = []
        basevalue = backports.entry_points_selectable.entry_points
        # Use the mocker only for 'numina.pipeline.1'

        def mockreturn(group, name=None):
            if group == 'numina.pipeline.1':
                return self._eps
            else:
                return basevalue(group=group, name=name)

        self.monkeypatch.setattr(
            backports.entry_points_selectable, 'entry_points', mockreturn)

    def add_drp(self, name, loader):

        if callable(loader):
            ep = create_mock_entry_point(self.monkeypatch, name, loader)
        else:
            # Assume loader is data instead
            drpdata = loader

            def drploader():
                return pload.drp_load_data('numina', drpdata)

            ep = create_mock_entry_point(self.monkeypatch, name, drploader)

        self._eps.append(ep)



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.units import Unit
import astropy.units as u
from astropy.visualization import ZScaleInterval
import matplotlib.pyplot as plt
import numpy as np

from numina.array.distortion import fmap
from numina.array.display.polfit_residuals import polfit_residuals
from numina.array.distortion import compute_distortion, rectify2d


def update_image2d_rss_method1(
        islice,
        image2d_detector_method0,
        dict_ifu2detector,
        naxis1_detector,
        naxis1_ifu,
        wv_crpix1, wv_crval1, wv_cdelt1,
        image2d_rss_method1,
        debug=False
):
    """Update the RSS image from the detector image.

    The function updates the following 2D array:
    - image2d_rss_method1,
    with the data of the slice 'islice' in the detector image.

    This function can be executed in parallel.

    Parameters
    ----------
    islice : int
        Slice number.
    image2d_detector_method0 : `~numpy.ndarray`
        2D array containing the detector image.
    dict_ifu2detector : dict
        A Python dictionary containing the 2D polynomials that allow
        to transform (X, Y, wavelength) coordinates in the IFU focal
        plane to (X, Y) coordinates in the detector.
    naxis1_detector : `~astropy.units.Quantity`
        Detector NAXIS1 (along the spectral direction).
    naxis1_ifu : `~astropy.units.Quantity`
        IFU NAXIS1 (along the slice).
    wv_crpix1 : `~astropy.units.Quantity`
        CRPIX1 value along the spectral direction.
    wv_crval1 : `~astropy.units.Quantity`
        CRVAL1 value along the spectral direction.
    wv_cdelt1 : `~astropy.units.Quantity`
        CDELT1 value along the spectral direction.
    image2d_rss_method1 : `~numpy.ndarray`
        2D array containing the RSS image. This array is
        updated by this function.
    debug : bool
        If True, show debugging information/plots.

    """

    slice_id = islice + 1

    # minimum and maximum pixel X coordinate defining the IFU focal plane
    min_x_ifu = 0.5 * u.pix
    max_x_ifu = naxis1_ifu + 0.5 * u.pix

    # determine upper and lower frontiers of each slice in the detector
    x_ifu_lower = np.repeat([min_x_ifu.value], naxis1_detector.value)
    x_ifu_upper = np.repeat([max_x_ifu.value], naxis1_detector.value)

    # wavelength values at each pixel in the spectral direction of the detector
    wavelength = wv_crval1 + ((np.arange(naxis1_detector.value) + 1) * u.pix - wv_crpix1) * wv_cdelt1

    # minimum and maximum wavelengths to be considered
    wmin = wv_crval1 + (0.5 * u.pix - wv_crpix1) * wv_cdelt1
    wmax = wmin + naxis1_detector * wv_cdelt1

    # use model to predict location in detector
    # important: reverse here X <-> Y
    wavelength_unit = Unit(dict_ifu2detector['wavelength-unit'])
    dumdict = dict_ifu2detector['contents'][islice]
    order = dumdict['order']
    aij = np.array(dumdict['aij'])
    bij = np.array(dumdict['bij'])

    y_detector_lower_index, x_detector_lower_index = fmap(
        order=order,
        aij=aij,
        bij=bij,
        x=x_ifu_lower,
        y=wavelength.to(wavelength_unit).value
    )
    # subtract 1 to work with array indices
    x_detector_lower_index -= 1
    y_detector_lower_index -= 1

    y_detector_upper_index, x_detector_upper_index = fmap(
        order=order,
        aij=aij,
        bij=bij,
        x=x_ifu_upper,
        y=wavelength.to(wavelength_unit).value
    )
    # subtract 1 to work with array indices
    x_detector_upper_index -= 1
    y_detector_upper_index -= 1

    if debug:
        debugplot = 1
    else:
        debugplot = 0
    poly_lower_index, residuals = polfit_residuals(
        x=x_detector_lower_index,
        y=y_detector_lower_index,
        deg=order,
        xlabel='x_detector_lower_index',
        ylabel='y_detector_lower_index',
        title=f'slice_id #{slice_id}',
        debugplot=debugplot
    )
    if debug:
        plt.tight_layout()
        plt.show()
    poly_upper_index, residuals = polfit_residuals(
        x=x_detector_upper_index,
        y=y_detector_upper_index,
        deg=order,
        xlabel='x_detector_upper_index',
        ylabel='y_detector_upper_index',
        title=f'slice_id #{slice_id}',
        debugplot=debugplot
    )
    if debug:
        plt.tight_layout()
        plt.show()

    # full image containing only the slice data (zero elsewhere)
    image2d_detector_slice = np.zeros_like(image2d_detector_method0)
    xdum = np.arange(naxis1_detector.value)
    ypoly_lower_index = (poly_lower_index(xdum) + 0.5).astype(int)
    ypoly_upper_index = (poly_upper_index(xdum) + 0.5).astype(int)
    for j in range(naxis1_detector.value):
        i1 = ypoly_lower_index[j]
        i2 = ypoly_upper_index[j]
        image2d_detector_slice[i1:(i2 + 1), j] = image2d_detector_method0[i1:(i2 + 1), j]

    if debug:
        xmin = np.min(np.concatenate((x_detector_lower_index, x_detector_upper_index)))
        xmax = np.max(np.concatenate((x_detector_lower_index, x_detector_upper_index)))
        ymin = np.min(np.concatenate((y_detector_lower_index, y_detector_upper_index)))
        ymax = np.max(np.concatenate((y_detector_lower_index, y_detector_upper_index)))
        print(f'{xmin=}, {xmax=}, {ymin=}, {ymax=}')
        dy = ymax - ymin
        yminplot = ymin - dy / 5
        ymaxplot = ymax + dy / 5
        fig, axarr = plt.subplots(nrows=2, ncols=1, figsize=(6.4*2, 4.8*2))
        vmin, vmax = ZScaleInterval().get_limits(image2d_detector_method0)
        for iplot in range(2):
            ax = axarr[iplot]
            if iplot == 0:
                ax.imshow(image2d_detector_method0, vmin=vmin, vmax=vmax, aspect='auto')
            else:
                ax.imshow(image2d_detector_slice, vmin=vmin, vmax=vmax, aspect='auto')
            ax.plot(x_detector_lower_index, y_detector_lower_index, 'C1--')
            ax.plot(x_detector_upper_index, y_detector_upper_index, 'C1--')
            ax.plot(xdum, ypoly_lower_index, 'w:')
            ax.plot(xdum, ypoly_upper_index, 'w:')
            ax.set_ylim(yminplot, ymaxplot)
        plt.tight_layout()
        plt.show()

    # mathematical transformation for the considered slice
    i1_rss_index = islice * naxis1_ifu.value
    i2_rss_index = i1_rss_index + naxis1_ifu.value
    if debug:
        print(f'{i1_rss_index=}, {i2_rss_index=}, {i2_rss_index - i1_rss_index=}')
    # generate a grid to compute the 2D transformation
    nx_grid_rss = 20
    ny_grid_rss = 20
    # points along the slice in the IFU focal pline
    x_ifu_grid = np.tile(np.linspace(1, naxis1_ifu.value, num=ny_grid_rss), nx_grid_rss)
    wavelength_grid = np.repeat(np.linspace(wmin, wmax, num=nx_grid_rss), ny_grid_rss)
    # pixel in the RSS image
    wavelength_grid_pixel_rss = (wavelength_grid - wv_crval1) / wv_cdelt1 + wv_crpix1
    if debug:
        fig, ax = plt.subplots(figsize=(6.4 * 2, 4.8))
        ax.plot(wavelength_grid_pixel_rss.value, x_ifu_grid + i1_rss_index, 'r.')
        ax.set_xlabel('RSS pixel in wavelength direction')
        ax.set_ylabel('RSS pixel in spatial direction')
        plt.tight_layout()
        plt.show()
    # project the previous points in the detector
    y_hawaii_grid_index, x_hawaii_grid_index = fmap(
        order=order,
        aij=aij,
        bij=bij,
        x=x_ifu_grid,
        y=wavelength_grid.to(wavelength_unit).value   # ignore PyCharm warning here
    )
    x_hawaii_grid_index -= 1
    y_hawaii_grid_index -= 1

    y_hawaii_grid_index, x_hawaii_grid_index = fmap(
        order=order,
        aij=aij,
        bij=bij,
        x=x_ifu_grid,
        y=wavelength_grid.to(wavelength_unit).value  # ignore PyCharm warning here
    )
    x_hawaii_grid_index -= 1
    y_hawaii_grid_index -= 1
    if debug:
        fig, ax = plt.subplots(figsize=(6.4 * 2, 4.8))
        ax.imshow(image2d_detector_slice, vmin=vmin, vmax=vmax, aspect='auto')  # ignore PyCharm warning here
        ax.plot(x_hawaii_grid_index, y_hawaii_grid_index, 'r.')
        ax.set_ylim(yminplot, ymaxplot)  # ignore PyCharm warning here
        plt.tight_layout()
        plt.show()
    # compute distortion transformation
    aij_resample, bij_resample = compute_distortion(
        x_orig=x_hawaii_grid_index + 1,
        y_orig=y_hawaii_grid_index + 1,
        x_rect=wavelength_grid_pixel_rss.value,
        y_rect=x_ifu_grid,
        order=order,
        debugplot=0
    )

    # rectify image
    image2d_slice_rss = rectify2d(
        image2d=image2d_detector_slice,
        aij=aij_resample,
        bij=bij_resample,
        resampling=2,  # 2: flux preserving interpolation
        naxis2out=naxis1_ifu.value
    )
    if debug:
        fig, ax = plt.subplots(figsize=(6.4 * 2, 4.8))
        ax.imshow(image2d_slice_rss, vmin=vmin, vmax=vmax, aspect='auto')
        plt.tight_layout()
        plt.show()

    # insert result in final image
    image2d_rss_method1[i1_rss_index:i2_rss_index, :] = image2d_slice_rss[:, :]




import pytest

from ..pipelineload import load_confs


@pytest.mark.xfail(reason="fix this later")
def test_load_confs_1():
    """Test the loader returns a valid config when 'values' is empty"""

    package = 'clodiadrp'
    node = {
        'values': {}
    }

    result = load_confs(package, node)
    assert len(result) == 2

    result1, result2 = result

    assert len(result1) == 1
    assert 'default' in result1

    insconf = result1['default']
    assert insconf.instrument == 'EMPTY'
    assert insconf.name == 'EMPTY'



import pytest

from ..device import DeviceBase
from ..state import Status


def test_device_base():
    dev = DeviceBase("dev1")

    assert dev.is_configured is False


def test_device_parent1():
    base = DeviceBase("base")

    other1 = DeviceBase("other1", parent=base)
    other2 = DeviceBase("other2", parent=base)

    acc1 = base.get_device("base.other1")
    acc2 = base.get_device("base.other2")

    assert acc1 is other1
    assert acc2 is other2


def test_device_parent2():

    base = DeviceBase("base")
    other1_1 = DeviceBase("other1", parent=base)  # noqa: F841
    other1_2 = DeviceBase("other1")

    # Duplicated name in tree
    with pytest.raises(ValueError):
        other1_2.set_parent(base)


def test_device_from():
    obj = DeviceBase.from_component("Base", "obj1")
    assert isinstance(obj, DeviceBase)
    assert obj.name == "obj1"
    assert obj.origin is None
    assert obj.parent is None


def test_device_set_parent():
    obj1 = DeviceBase("obj1")
    obj2 = DeviceBase("obj2", parent=obj1)
    obj1b = DeviceBase("obj1b")
    assert len(obj1b.children) == 0
    assert obj1.children["obj2"] is obj2
    assert obj2.parent is obj1
    obj2.set_parent(obj1b)
    assert obj2.parent is obj1b
    assert obj1b.children["obj2"] is obj2
    assert len(obj1.children) == 0


def test_device_get_value():
    obj1 = DeviceBase("obj1")
    u = obj1.get_value("name")
    assert u == "obj1"


def test_configure_image():
    import astropy.io.fits as fits

    test_hdu = fits.PrimaryHDU()
    test_img = fits.HDUList([test_hdu])
    obj1 = DeviceBase("obj1")
    obj2 = DeviceBase("obj2", parent=obj1)  # noqa
    obj1.configure_with_image(test_img)
    assert obj1.state.current == Status.ACTIVE


def test_property_names():
    obj1 = DeviceBase("obj1")
    assert obj1.get_property_names() == set([])


def test_get_properties():
    obj1 = DeviceBase("obj1")
    assert obj1.get_properties(init=False) == {}


def test_end():
    obj1 = DeviceBase("obj1")
    obj2 = DeviceBase("obj2", parent=obj1)
    res1 = obj1.end_config_info({})
    assert res1 == {"children": ["obj2"]}
    res2 = obj2.end_config_info({})
    assert res2 == {}



import pytest

from numina.exceptions import ValidationError
from numina.types.datatype import ListOfType, PlainPythonType


@pytest.mark.parametrize("nmin, nmax, allowed, not_allowed", [
    (None, None, [[], [1, 2, 3], [-1, 2.0, 34.0, 4, -5]], []),
    (1, None, [[-1, 2.0, 34.0, 4, -5]], [[]]),
    (None, 3, [[], [1], [2, 3], [1, 2, 3]], [[-1, 2.0, 34.0, 4, -5]]),
    (4, 4, [[-1, 2.0, 34.0, 4]], [[1, 2, 3], []]),
])
def test_list(nmin, nmax, allowed, not_allowed):
    ppt = ListOfType(PlainPythonType(ref=0), nmin=nmin, nmax=nmax)

    for obj1 in allowed:
        assert ppt.convert(obj1) == obj1

    for obj2 in not_allowed:
        with pytest.raises(ValidationError):
            ppt.convert(obj2)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Resample a 3D cube in the wavelength axis (NAXIS3).
"""
import sys

import argparse
from astropy.io import fits
import astropy.units as u
from astropy.wcs import WCS
import numpy as np

from numina.instrument.simulation.ifu.define_3d_wcs \
    import header3d_after_merging_wcs2d_celestial_and_wcs1d_spectral

from .add_script_info_to_fits_history import add_script_info_to_fits_history
from .ctext import ctext


def resample_wave_3d_cube(hdu3d_image, crval3out, cdelt3out, naxis3out, verbose=False):
    """Resample a 3D cube to a new wavelength sampling.

    The celestial WCS is preserved, and the spectral WCS is modified
    in order to make use of the new wavelength sampling.

    Parameters
    ----------
    hdu3d_image : `astropy.io.fits.ImageHDU`
        HDU instance with the 3D image to be resampled.
    crval3out : `astropy.units.Quantity`
        Minimum wavelength for the output image.
    cdelt3out : `astropy.units.Quantity`
        Wavelength step for the output image.
    naxis3out : int
        Number of slices in the output image.
    verbose : bool, optional
        If True, display intermediate information.

    Returns
    -------
    resampled_hdu : `astropy.io.fits.ImageHDU`
        Resampled HDU instance with the 3D image.
    """
    # protections
    if not isinstance(hdu3d_image, fits.ImageHDU) and not isinstance(hdu3d_image, fits.PrimaryHDU):
        raise ValueError("Input HDU must be an ImageHDU or PrimaryHDU.")
    if crval3out is None:
        raise ValueError("crval3out must be specified.")
    if cdelt3out is None:
        raise ValueError("cdelt3out must be specified.")
    if naxis3out is None:
        raise ValueError("naxis3out must be specified.")
    if hdu3d_image.data.ndim != 3:
        raise ValueError("Input HDU must be a 3D cube.")

    # get shape of the input 3D cube
    naxis3, naxis2, naxis1 = hdu3d_image.data.shape

    # initial pixel borders in the spectral axis
    old_wcs1d_spectral = WCS(hdu3d_image.header).spectral
    old_wl_borders = old_wcs1d_spectral.pixel_to_world(np.arange(naxis3+1)-0.5)
    # modify slightly the first and last values to avoid numerical issues
    deltawave = old_wl_borders[1] - old_wl_borders[0]
    old_wl_borders[0] = old_wl_borders[0] - deltawave/1E6
    deltawave = old_wl_borders[-1] - old_wl_borders[-2]
    old_wl_borders[-1] = old_wl_borders[-1] + deltawave/1E6

    # final pixel borders in the spectral axis
    new_wl_borders = crval3out + cdelt3out * (np.arange(naxis3out + 1) - 0.5) * u.pix

    resample_needed = True
    if naxis3 == naxis3out:
        # if the old and new wavelength borders are the same, just copy the data
        if np.all(np.allclose(old_wl_borders, new_wl_borders)):
            resampled_data = hdu3d_image.data.astype(np.float32)
            resample_needed = False
            if verbose:
                print("Old and new wavelength borders are the same.\n"
                      "-> Copying original data without spectral resampling.")

    if resample_needed:
        if verbose:
            print("Spectral resampling of the original 3D cube...", end=' ')
        # resample the 3D cube (see wavecal.py in teareduce for reference)
        resampled_data = np.zeros((naxis3out, naxis2, naxis1))
        for i in range(naxis1):
            for j in range(naxis2):
                # resample each spectrum independently
                data_spectrum = hdu3d_image.data[:, j, i]
                accum_flux = np.zeros(naxis3 + 1)
                accum_flux[1:] = np.cumsum(data_spectrum)
                flux_borders = np.interp(
                    x=new_wl_borders.value,
                    xp=old_wl_borders.value,
                    fp=accum_flux,
                    left=np.nan,
                    right=np.nan
                )
                resampled_data[:, j, i] = flux_borders[1:] - flux_borders[:-1]
        if verbose:
            print("resampling completed!")

    # create new HDU with resampled data
    resampled_hdu = fits.PrimaryHDU(data=resampled_data.astype(np.float32))
    header_spectral_resampled = fits.Header()
    header_spectral_resampled['NAXIS'] = 1
    header_spectral_resampled['NAXIS1'] = naxis3out
    header_spectral_resampled['CRPIX1'] = 1.0
    header_spectral_resampled['CDELT1'] = cdelt3out.to(u.m/u.pix).value
    header_spectral_resampled['CRVAL1'] = crval3out.to(u.m).value
    header_spectral_resampled['CUNIT1'] = 'm'
    header_spectral_resampled['CTYPE1'] = 'WAVE'
    wcs1d_spectral_resampled = WCS(header_spectral_resampled)
    header_resampled = header3d_after_merging_wcs2d_celestial_and_wcs1d_spectral(
        wcs2d_celestial=WCS(hdu3d_image.header).celestial,
        wcs1d_spectral=wcs1d_spectral_resampled
    )
    resampled_hdu.header.update(header_resampled)

    return resampled_hdu


def main(args=None):
    """Main function."""
    parser = argparse.ArgumentParser(
        description="Resample a 3D cube in the wavelength axis (NAXIS3)."
    )
    parser.add_argument("input_file", type=str,
                        help="Input FITS file with the 3D cube.")
    parser.add_argument("output_file", type=str,
                        help="Output FITS file with the resampled 3D cube.")
    parser.add_argument("--crval3out", type=float,
                        help="Minimum wavelength for the output image (in meters).")
    parser.add_argument("--cdelt3out", type=float,
                        help="Wavelength step for the output image (in meters).")
    parser.add_argument("--naxis3out", type=int,
                        help="Number of slices in the output image.")
    parser.add_argument("--extname", type=str,
                        help="Extension name of the input HDU (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--verbose",
                        help="Display intermediate information",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    input_file = args.input_file
    output_file = args.output_file
    crval3out = args.crval3out
    if crval3out is not None:
        crval3out = crval3out * u.m
    cdelt3out = args.cdelt3out
    if cdelt3out is not None:
        cdelt3out = cdelt3out * u.m / u.pix
    naxis3out = args.naxis3out
    extname = args.extname
    verbose = args.verbose

    with fits.open(input_file) as hdul:
        if extname not in hdul:
            raise ValueError(f"Extension '{extname}' not found in {input_file}.")
        hdu3d_image = hdul[extname].copy()
        if verbose:
            print(f"{hdu3d_image.header['NAXIS1']=}")
            print(f"{hdu3d_image.header['NAXIS2']=}")
            print(f"{hdu3d_image.header['NAXIS3']=}")

    if crval3out is None or cdelt3out is None:
        wcs1d_spectral = WCS(hdu3d_image.header).spectral
        wave = wcs1d_spectral.pixel_to_world(np.arange(hdu3d_image.data.shape[0]))
        if crval3out is None:
            crval3out = wave[0]
            if verbose:
                print(f"Assuming {crval3out=}.")
        if cdelt3out is None:
            cdelt3out = (wave[1] - wave[0]) / u.pix
            if verbose:
                print(f"Assuming {cdelt3out=}.")

    if naxis3out is None:
        naxis3out = hdu3d_image.data.shape[0]
        if verbose:
            print(f"Assuming {naxis3out=}.")

    resampled_hdu = resample_wave_3d_cube(
        hdu3d_image=hdu3d_image,
        crval3out=crval3out,
        cdelt3out=cdelt3out,
        naxis3out=naxis3out,
        verbose=verbose
    )

    add_script_info_to_fits_history(resampled_hdu.header, args)
    if verbose:
        print(f'Saving: {output_file}')
    resampled_hdu.writeto(output_file, overwrite=True)


if __name__ == "__main__":
    main()



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Include ADR prediction as extension in a 3D FITS file"""

import argparse
from astropy.io import fits
import astropy.units as u
from astropy.units import Unit
from astropy.coordinates import SkyCoord, Angle
from astropy.wcs import WCS
import numpy as np
import sys

from .compute_adr_wavelength import compute_adr_wavelength
from .compare_adr_extensions_in_3d_cube import compare_adr_extensions_in_3d_cube
from .ctext import ctext


def include_adrtheor_in_3d_cube(
        filename,
        extname,
        reference_vacuum_wavelength_angstrom,
        temperature,
        pressure_mm,
        pressure_water_vapor_mm,
        plots,
        verbose=False
):
    """Include ADR prediction as extension in a 3D FITS file

    Parameters
    ----------
    filename : str, file-like or `pathlib.Path`
        FITS filename to be updated.
    extname : str
        Extension to store the ADR prediction in.
    reference_vacuum_wavelength_angstrom : float or None
        Reference wavelength (in Angstrom) to compute ADR prediction.
    verbose : bool
        If True, display additional information.
    """

    with fits.open(filename) as hdul:
        if verbose:
            print(hdul.info())
        header = hdul[0].header

    if header["NAXIS"] != 3:
        raise ValueError("Expected NAXIS=3 not found in PRIMARY HDU")

    # generate 3D WCS object
    wcs3d = WCS(header)
    naxis1, naxis2, naxis3 = wcs3d.pixel_shape
    if verbose:
        print(wcs3d.spectral)

    # array of wavelengths along NAXIS3
    wave = wcs3d.spectral.pixel_to_world(np.arange(naxis3))

    # reference wavelength to compute ADR
    if reference_vacuum_wavelength_angstrom is not None:
        reference_vacuum_wavelength = reference_vacuum_wavelength_angstrom * Unit('Angstrom')
        reference_vacuum_wavelength = reference_vacuum_wavelength.to(Unit('m'))
    else:
        reference_vacuum_wavelength = (wave[0] + wave[-1]) / 2
    if verbose:
        print(f'Reference wavelength: {reference_vacuum_wavelength}')

    # airmass
    if 'AIRMASS' in header:
        airmass = header['AIRMASS']
        if verbose:
            print(f'AIRMASS: {airmass}')
    else:
        raise ValueError('Header does not contain AIRMASS')

    # differential refraction (arcsec)
    differential_refraction = compute_adr_wavelength(
        airmass=airmass,
        reference_wave_vacuum=reference_vacuum_wavelength,
        wave_vacuum=wave,
        temperature=temperature,
        pressure_mm=pressure_mm,
        pressure_water_vapor_mm=pressure_water_vapor_mm,
    )
    if verbose:
        print(f'Differential refraction: {differential_refraction}')

    # parallactic angle
    if 'PARANGLE' in header:
        parangle = Angle(header['PARANGLE'] * Unit('deg'))
        if verbose:
            print(f'PARANGLE: {parangle}')
    else:
        raise ValueError('Header does not contain PARANGLE')

    # predict ADR at the center of the field of view
    x_center_ifu, y_center_ifu = wcs3d.celestial.wcs.crpix    # FITS convention
    center_ifu_coord = wcs3d.celestial.pixel_to_world(
        x_center_ifu - 1.0,   # Python convention
        y_center_ifu - 1.0    # Python convention
    )
    if verbose:
        print(f'Center IFU coord: {center_ifu_coord}')

    # duplicate initial central coordinates at each slice along NAXIS3
    ra_center_ifu = np.repeat(center_ifu_coord.ra, naxis3)
    dec_center_ifu = np.repeat(center_ifu_coord.dec, naxis3)

    # apply differential refraction correction
    dec_center_ifu += differential_refraction.to(u.deg) * np.cos(parangle)
    ra_center_ifu += differential_refraction.to(u.deg) * np.sin(parangle) / np.cos(dec_center_ifu)

    # compute variation in (X,Y) coordinates
    x_center_ifu_corrected, y_center_ifu_corrected = wcs3d.celestial.world_to_pixel(
        SkyCoord(ra=ra_center_ifu, dec=dec_center_ifu)
    )
    x_center_ifu_corrected += 1
    y_center_ifu_corrected += 1
    delta_x_center_ifu = x_center_ifu_corrected - x_center_ifu
    delta_y_center_ifu = y_center_ifu_corrected - y_center_ifu

    # save result in extension
    if extname != 'NONE':
        if verbose:
            print(f'Updating file {filename}')
        # binary table to store result
        col1 = fits.Column(name='Delta_x', format='D', array=delta_x_center_ifu, unit='pixel')
        col2 = fits.Column(name='Delta_y', format='D', array=delta_y_center_ifu, unit='pixel')
        hdu_result = fits.BinTableHDU.from_columns([col1, col2])
        hdu_result.name = extname.upper()
        hdu_result.header['AIRMASS'] = (airmass, 'Airmass')
        hdu_result.header['PARANGLE'] = (parangle.value, 'Parallactic angle (deg)')
        hdu_result.header['REFEWAVE'] = (reference_vacuum_wavelength.to(u.m).value,
                                         'Reference vacuum wavelength (m)')
        hdu_result.header['TEMPERAT'] = (temperature.value, 'Assumed temperature (Celsius degrees)')
        hdu_result.header['PRESSURE'] = (pressure_mm, 'Pressure (Hg mm)')
        hdu_result.header['PRESSUWV'] = (pressure_water_vapor_mm, 'Water vapor pressure (Hg mm)')
        # open and update existing FITS file
        hdul = fits.open(filename, mode='update')
        if extname in hdul:
            if verbose:
                print(f"Updating extension '{extname}'")
            hdul[extname] = hdu_result
        else:
            if verbose:
                print(f"Adding new extension '{extname}'")
            hdul.append(hdu_result)
        hdul.flush()
        hdul.close()

    # display results
    if plots:
        compare_adr_extensions_in_3d_cube(filename, extname1=extname, extname2=None)


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(description="Include ADR prediction as extension in a 3D FITS file")
    parser.add_argument("filename", help="Input 3D FITS file")
    parser.add_argument("--extname",
                        help="Output extension name to store result (default ADRTHEOR)",
                        type=str, default='ADRTHEOR')
    parser.add_argument("--reference_vacuum_wavelength",
                        help="Reference vacuum wavelength (in Angstrom) to compute ADR prediction",
                        type=float, default=None)
    parser.add_argument("--temperature", help="Temperature in degree Celsius",
                        type=float, default=7)
    parser.add_argument("--pressure_mm", help="Pressure in Hg mm",
                        type=float, default=600)
    parser.add_argument("--pressure_water_vapor_mm", help="Pressure water vapor in Hg mm",
                        type=float, default=8)
    parser.add_argument("--plots", help="Plot intermediate results", action="store_true")
    parser.add_argument("--verbose", help="Display intermediate information", action="store_true")
    parser.add_argument("--echo", help="Display full command line", action="store_true")

    args = parser.parse_args(args=args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # protections
    extname = args.extname.upper()
    if len(extname) > 8:
        raise ValueError(f"Extension '{extname}' must be less than 9 characters")

    reference_vacuum_wavelength_angstrom = args.reference_vacuum_wavelength
    if reference_vacuum_wavelength_angstrom is not None:
        if reference_vacuum_wavelength_angstrom <= 0.0:
            raise ValueError("Reference vacuum wavelength must be positive")

    include_adrtheor_in_3d_cube(
        filename=args.filename,
        extname=extname,
        reference_vacuum_wavelength_angstrom=reference_vacuum_wavelength_angstrom,
        temperature=args.temperature*u.Celsius,
        pressure_mm=args.pressure_mm,
        pressure_water_vapor_mm=args.pressure_water_vapor_mm,
        plots=args.plots,
        verbose=args.verbose
    )


if __name__ == '__main__':

    main()



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from itertools import product

from numpy.lib.stride_tricks import as_strided as ast


def blockgen1d(block, size):
    """Compute 1d block intervals to be used by combine.

    blockgen1d computes the slices by recursively halving the initial
    interval (0, size) by 2 until its size is lesser or equal than block

    :param block: an integer maximum block size
    :param size: original size of the interval, it corresponds to a 0:size slice
    :return: a list of slices

    Example:

        >>> blockgen1d(512, 1024)
        [slice(0, 512, None), slice(512, 1024, None)]

    """
    def numblock(blk, x):
        """Compute recursively the numeric intervals

        """
        a, b = x
        if b - a <= blk:
            return [x]
        else:
            result = []
            d = int(b - a) // 2
            for i in map(numblock, [blk, blk], [(a, a + d), (a + d, b)]):
                result.extend(i)
            return result

    return [slice(*ll) for ll in numblock(block, (0, size))]


def blockgen(blocks, shape):
    """Generate a list of slice tuples to be used by combine.

    The tuples represent regions in an N-dimensional image.

    :param blocks: a tuple of block sizes
    :param shape: the shape of the n-dimensional array
    :return: an iterator to the list of tuples of slices

    Example:

        >>> iblocks = (500, 512)
        >>> ishape = (1040, 1024)
        >>> for reg in blockgen(iblocks, ishape):
        ...     print(reg)
        (slice(0, 260, None), slice(0, 512, None))
        (slice(0, 260, None), slice(512, 1024, None))
        (slice(260, 520, None), slice(0, 512, None))
        (slice(260, 520, None), slice(512, 1024, None))
        (slice(520, 780, None), slice(0, 512, None))
        (slice(520, 780, None), slice(512, 1024, None))
        (slice(780, 1040, None), slice(0, 512, None))
        (slice(780, 1040, None), slice(512, 1024, None))

    """
    iterables = [blockgen1d(l, s) for (l, s) in zip(blocks, shape)]
    return product(*iterables)


def blk_coverage_1d(blk, size):
    """Return the part of a 1d array covered by a block.

    :param blk: size of the 1d block
    :param size: size of the 1d a image
    :return: a tuple of size covered and remaining size

    Example:

        >>> blk_coverage_1d(7, 100)
        (98, 2)

    """
    rem = size % blk
    maxpix = size - rem
    return maxpix, rem


def max_blk_coverage(blk, shape):
    """Return the maximum shape of an array covered by a block.

    :param blk: the N-dimensional shape of the block
    :param shape: the N-dimensional shape of the array
    :return: the shape of the covered region

    Example:

        >>> max_blk_coverage(blk=(7, 6), shape=(100, 43))
        (98, 42)


    """
    return tuple(blk_coverage_1d(b, s)[0] for b, s in zip(blk, shape))


def blk_1d_short(blk, shape):
    """Iterate through the slices that recover a line.

    This function is used by :func:`blk_nd_short` as a base 1d case.

    The function stops yielding slices when the size of
    the remaining slice is lesser than `blk`.

    :param blk: the size of the block
    :param shape: the size of the array
    :return: a generator that yields the slices
    """
    maxpix, _ = blk_coverage_1d(blk, shape)
    for i in range(0, maxpix, blk):
        yield slice(i, i + blk)


def blk_nd_short(blk, shape):
    """Iterate trough the blocks that strictly cover an array.

    Iterate trough the blocks that recover the part of the array
    given by max_blk_coverage.

    :param blk: the N-dimensional shape of the block
    :param shape: the N-dimensional shape of the array
    :return: a generator that yields the blocks

    Example:

        >>> result = list(blk_nd_short(blk=(5,3), shape=(11, 11)))
        >>> result[0]
        (slice(0, 5, None), slice(0, 3, None))
        >>> result[1]
        (slice(0, 5, None), slice(3, 6, None))
        >>> result[-1]
        (slice(5, 10, None), slice(6, 9, None))

        In this case, the output of max_blk_coverage
        is (10, 9), so only this part of the array is covered


    .. seealso::

        :py:func:`blk_nd`
          Yields blocks of blk size until the remaining part is
          smaller than `blk` and the yields smaller blocks.

    """
    internals = (blk_1d_short(b, s) for b, s in zip(blk, shape))
    return product(*internals)


def blk_1d(blk, shape):
    """Iterate through the slices that recover a line.

    This function is used by :func:`blk_nd` as a base 1d case.

    The last slice  is returned even if is lesser than blk.

    :param blk: the size of the block
    :param shape: the size of the array
    :return: a generator that yields the slices

    """
    maxpix, rem = blk_coverage_1d(blk, shape)
    for i in range(0, maxpix, blk):
        yield slice(i, i + blk)

    if rem != 0:
        yield slice(maxpix, shape)


def blk_nd(blk, shape):
    """Iterate through the blocks that cover an array.

    This function first iterates trough the blocks that recover
    the part of the array given by max_blk_coverage
    and then iterates with smaller blocks for the rest
    of the array.

    :param blk: the N-dimensional shape of the block
    :param shape: the N-dimensional shape of the array
    :return: a generator that yields the blocks

    Example:

        >>> result = list(blk_nd(blk=(5,3), shape=(11, 11)))
        >>> result[0]
        (slice(0, 5, None), slice(0, 3, None))
        >>> result[1]
        (slice(0, 5, None), slice(3, 6, None))
        >>> result[-1]
        (slice(10, 11, None), slice(9, 11, None))

    The generator yields blocks of size blk until
    it covers the part of the array given by
    :func:`max_blk_coverage` and then yields
    smaller blocks until it covers the full array.

    .. seealso::

        :py:func:`blk_nd_short`
          Yields blocks of fixed size

    """
    internals = (blk_1d(b, s) for b, s in zip(blk, shape))
    return product(*internals)


def block_view(arr, block=(3, 3)):
    """Provide a 2D block view to 2D array.

    No error checking made. Therefore meaningful (as implemented) only for
    blocks strictly compatible with the shape of A.

    """

    # simple shape and strides computations may seem at first strange
    # unless one is able to recognize the 'tuple additions' involved ;-)
    shape = (arr.shape[0] // block[0], arr.shape[1] // block[1]) + block
    strides = (block[0] * arr.strides[0], block[1]
               * arr.strides[1]) + arr.strides
    return ast(arr, shape=shape, strides=strides)



def test_configure1(pattri_ins, pattri_state):
    pattri_ins.configure(pattri_state)
    assert pattri_ins.get_device("wheel").label == "U"
    assert pattri_ins.get_property("wheel.position") == 0
    assert pattri_ins.get_property("robot.arm_2.angle") == 1.2


def test_configure2(pattri_ins, pattri_state):
    pattri_ins.configure(pattri_state)

    res = pattri_ins.config_info()
    assert "PATTRI.wheel" in res


def test_configure3(pattri_ins, pattri_state):
    pattri_ins.configure(pattri_state)

    res = pattri_ins.config_info()
    assert "PATTRI.detector" in res
    assert pattri_ins.get_device("detector").shape == [2048, 2048]
    assert res["PATTRI.detector"] == {}


def test_configure4(pattri_ins, pattri_state):
    pattri_ins.configure(pattri_state)
    info = pattri_ins.config_info()
    assert info["PATTRI.PSU.BarL_1"] == {"pos": -8}
    assert pattri_ins.get_device("PATTRI.PSU.BarL_1").pos == -8
    assert pattri_ins.get_property("PATTRI.PSU.BarL_1.pos") == -8



#
# Copyright 2014-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DAL base class"""

from abc import ABCMeta, abstractmethod


class DALInterface(metaclass=ABCMeta):
    """Abstract Base Class for DAL"""

    @abstractmethod
    def obsres_from_oblock_id(self, obsid):
        # This implementation does not depend
        # on the details of the DAL...
        pass

    @abstractmethod
    def search_recipe_from_ob(self, ob, pipeline='default'):
        # returns RecipeClass
        pass

    @abstractmethod
    def search_product(self, name, tipo, obsres, options=None):
        # returns StoredProduct
        pass

    @abstractmethod
    def search_parameter(self, name, tipo, obsres, options=None):
        # returns StoredProduct
        pass

    @abstractmethod
    def search_result_relative(self, name, tipo, obsres, result_desc, options=None):
        pass



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""FWHM calculation"""

import numpy as np
import scipy.interpolate as itpl

from numina.array.utils import coor_to_pix_1d


def compute_fwhm_2d_simple(img, xc, yc):

    X = np.arange(0, img.shape[1], 1.0)
    Y = np.arange(0, img.shape[0], 1.0)

    xpix = coor_to_pix_1d(xc - X[0])
    ypix = coor_to_pix_1d(yc - Y[0])

    peak = img[ypix, xpix]

    res11 = img[ypix, :]
    res22 = img[:, xpix]

    fwhm_x, _codex, _msgx = compute_fwhm_1d(X, res11 - 0.5 * peak, xc, xpix)
    fwhm_y, _codey, _msgy = compute_fwhm_1d(Y, res22 - 0.5 * peak, yc, ypix)

    return peak, fwhm_x, fwhm_y


def compute_fwhm_2d_spline(img, xc, yc):

    Y = np.arange(0.0, img.shape[1], 1.0)
    X = np.arange(0.0, img.shape[0], 1.0)

    xpix = coor_to_pix_1d(xc)
    ypix = coor_to_pix_1d(yc)
    # The image is already cropped

    bb = itpl.RectBivariateSpline(X, Y, img)
    # We assume that the peak is in the center...
    peak = bb(xc, yc)[0, 0]

    U = X
    V = bb.ev(U, [yc for _ in U]) - 0.5 * peak
    fwhm_x, _codex, _msgx = compute_fwhm_1d(U, V, yc, ypix)

    U = Y
    V = bb.ev([xc for _ in U], U) - 0.5 * peak
    fwhm_y, _codey, _msgy = compute_fwhm_1d(U, V, xc, xpix)

    return peak, fwhm_x, fwhm_y


def compute_fwhm_1d_simple(Y, xc, X=None):
    """Compute the FWHM."""
    return compute_fw_at_frac_max_1d_simple(Y, xc, X=X, f=0.5)


def compute_fw_at_frac_max_1d_simple(Y, xc, X=None, f=0.5):
    """Compute the full width at fraction f of the maximum"""

    yy = np.asarray(Y)

    if yy.ndim != 1:
        raise ValueError('array must be 1-d')

    if yy.size == 0:
        raise ValueError('array is empty')

    if X is None:
        xx = np.arange(yy.shape[0])
    else:
        xx = X

    xpix = coor_to_pix_1d(xc - xx[0])

    try:
        peak = yy[xpix]
    except IndexError:
        raise ValueError('peak is out of array')

    fwhm_x, _codex, _msgx = compute_fwhm_1d(xx, yy - f * peak, xc, xpix)
    return peak, fwhm_x


def _fwhm_side_lineal(uu, vv):
    '''Compute r12 using linear interpolation.'''
    res1, = np.nonzero(vv < 0)
    if len(res1) == 0:
        return 0, 1  # error, no negative value
    else:
        # first value
        i2 = res1[0]
        i1 = i2 - 1
        dx = uu[i2] - uu[i1]
        dy = vv[i2] - vv[i1]
        r12 = uu[i1] - vv[i1] * dx / dy
        return r12, 0


def compute_fwhm_1d(uu, vv, uc, upix):

    _fwhm_side = _fwhm_side_lineal

    # Find half peak radius on the rigth
    r12p, errorp = _fwhm_side(uu[upix:], vv[upix:])

    # Find half peak radius on the left
    r12m, errorm = _fwhm_side(uu[upix::-1], vv[upix::-1])

    if errorm == 1:
        if errorp == 1:
            fwhm = -99  # No way
            msg = 'Failed to compute FWHM'
            code = 2
        else:
            fwhm = 2 * (r12p - uc)
            code = 1
            msg = 'FWHM computed from right zero'
    else:
        if errorp == 1:
            fwhm = 2 * (uc - r12m)
            msg = 'FWHM computed from left zero'
            code = 1
        else:
            msg = 'FWHM computed from left and right zero'
            code = 0
            fwhm = r12p - r12m

    return fwhm, code, msg



import pytest

from numina.exceptions import ValidationError
from numina.types.datatype import PlainPythonType


def validator_fun1(value):
    if value > 50:
        raise ValidationError
    return value


def test_no_check():
    value = 1
    obj = 2

    ppt = PlainPythonType(ref=value)
    assert ppt.convert(obj) == obj


def test_custom_check1():
    value = 1
    obj = 2

    ppt = PlainPythonType(ref=value, validator=validator_fun1)
    assert ppt.convert(obj) == obj


def test_custom_check2():
    value = 1
    obj = 200

    ppt = PlainPythonType(ref=value, validator=validator_fun1)
    with pytest.raises(ValidationError):
        ppt.convert(obj)


def test_conversion():
    value = 1.0
    obj1 = 2
    obj2 = 2.0

    ppt = PlainPythonType(ref=value, validator=validator_fun1)

    assert ppt.convert(obj1) == obj2
    assert isinstance(obj2, ppt.internal_type)
    assert ppt.internal_type is type(value)



#
# Copyright 2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Tests for alias in RecipeInput"""

from ..dataholders import Parameter
from ..recipeinout import RecipeInput


def create_input_class():

    class BB(RecipeInput):
        param1 = Parameter(1, 'something1', alias='param3')
        param2 = Parameter(2, 'something2')

    return BB


def test_ins_desc_access():

    BB = create_input_class()

    bb = BB(param1=10)

    assert bb.param1 == 10
    assert bb.param3 == 10


def test_ins_desc_access2():

    BB = create_input_class()

    bb = BB(param3=80)

    assert bb.param1 == 80
    assert bb.param3 == 80


def test_ins_attr_access():

    BB = create_input_class()

    bb = BB(param1=80)

    values = {'param2': 2, 'param1': 80}

    for key, val in bb.attrs().items():
        assert val == values[key]


def test_setter1():

    BB = create_input_class()

    bb = BB()
    bb.param1 = 80

    values = {'param1': 80, 'param2': 2, 'param3': 80}

    for key, val in values.items():
        assert val == getattr(bb, key)


def test_setter2():

    BB = create_input_class()

    bb = BB()
    bb.param3 = 80

    values = {'param1': 80, 'param2': 2, 'param3': 80}

    for key, val in bb.attrs().items():
        assert val == values[key]



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import numpy
import numpy as np


def robust_std(x, debug=False):
    """Compute a robust estimator of the standard deviation

    See Eq. 3.36 (page 84) in Statistics, Data Mining, and Machine
    in Astronomy, by Ivezic, Connolly, VanderPlas & Gray

    Parameters
    ----------
    x : 1d numpy array, float
        Array of input values which standard deviation is requested.
    debug : bool
        If True prints computed values

    Returns
    -------
    sigmag : float
        Robust estimator of the standar deviation
    """

    x = numpy.asarray(x)

    # compute percentiles and robust estimator
    q25 = numpy.percentile(x, 25)
    q75 = numpy.percentile(x, 75)
    sigmag = 0.7413 * (q75 - q25)

    if debug:
        print('debug|sigmag -> q25......................:', q25)
        print('debug|sigmag -> q75......................:', q75)
        print('debug|sigmag -> Robust standard deviation:', sigmag)

    return sigmag


def summary(x, rm_nan=False, debug=False):
    """Compute basic statistical parameters.

    Parameters
    ----------
    x : 1d numpy array, float
        Input array with values which statistical properties are
        requested.
    rm_nan : bool
        If True, filter out NaN values before computing statistics.
    debug : bool
        If True prints computed values.

    Returns
    -------
    result : Python dictionary
        Number of points, minimum, percentile 25, percentile 50
        (median), mean, percentile 75, maximum, standard deviation,
        robust standard deviation, percentile 15.866 (equivalent
        to -1 sigma in a normal distribution) and percentile 84.134
        (+1 sigma).

    """

    # protections
    if isinstance(x, np.ndarray):
        xx = np.copy(x)
    else:
        if isinstance(x, list):
            xx = np.array(x)
        else:
            raise ValueError('x=' + str(x) + ' must be a numpy.ndarray')

    if xx.ndim != 1:
        raise ValueError('xx.dim=' + str(xx.ndim) + ' must be 1')

    # filter out NaN's
    if rm_nan:
        xx = xx[np.logical_not(np.isnan(xx))]

    # compute basic statistics
    npoints = len(xx)
    ok = npoints > 0
    result = {
        'npoints': npoints,
        'minimum': np.min(xx) if ok else 0,
        'percentile25': np.percentile(xx, 25) if ok else 0,
        'median': np.percentile(xx, 50) if ok else 0,
        'mean': np.mean(xx) if ok else 0,
        'percentile75': np.percentile(xx, 75) if ok else 0,
        'maximum': np.max(xx) if ok else 0,
        'std': np.std(xx) if ok else 0,
        'robust_std': robust_std(xx) if ok else 0,
        'percentile15': np.percentile(xx, 15.86553) if ok else 0,
        'percentile84': np.percentile(xx, 84.13447) if ok else 0
    }

    if debug:
        print('>>> ========================================')
        print('>>> STATISTICAL SUMMARY:')
        print('>>> ----------------------------------------')
        print('>>> Number of points.........:', result['npoints'])
        print('>>> Minimum..................:', result['minimum'])
        print('>>> 1st Quartile.............:', result['percentile25'])
        print('>>> Median...................:', result['median'])
        print('>>> Mean.....................:', result['mean'])
        print('>>> 3rd Quartile.............:', result['percentile75'])
        print('>>> Maximum..................:', result['maximum'])
        print('>>> ----------------------------------------')
        print('>>> Standard deviation.......:', result['std'])
        print('>>> Robust standard deviation:', result['robust_std'])
        print('>>> 0.1586553 percentile.....:', result['percentile15'])
        print('>>> 0.8413447 percentile.....:', result['percentile84'])
        print('>>> ========================================')

    return result



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import datetime
import logging
import warnings

from astropy.io import fits

import numina.array as array
import numina.util.node as node
import numina.datamodel as dm


_logger = logging.getLogger(__name__)


def promote_hdulist(hdulist, totype='float32'):
    nn = [promote_hdu(hdu, totype=totype) for hdu in hdulist]
    newhdulist = fits.HDUList(nn)
    return newhdulist


def promote_hdu(hdu, totype='float32'):
    if hdu.data is None:
        return hdu
    newdata = hdu.data.astype(totype)
    newheader = hdu.header.copy()
    if isinstance(hdu, fits.PrimaryHDU):
        hdu = fits.PrimaryHDU(newdata, header=newheader)
        return hdu
    elif isinstance(hdu, fits.ImageHDU):
        hdu = fits.ImageHDU(newdata, header=newheader)
        return hdu
    else:
        # do nothing
        pass
    return hdu


class SimpleDataModel(dm.DataModel):
    """Model of the Data being processed"""
    pass


class Corrector(node.Node):
    """A Node that corrects a frame from instrumental signatures."""

    def __init__(self, datamodel=None, calibid='calibid-unknown', dtype='float32'):
        super(Corrector, self).__init__()
        if not datamodel:
            self.datamodel = SimpleDataModel()
        else:
            self.datamodel = datamodel
        self.dtype = dtype
        self.calibid = calibid

    def __call__(self, img):
        if img[0].data.dtype in ['<u2', '>u2', '=u2']:
            # FIXME: this is a GCS problem
            _logger.debug(
                'change dtype to float32, old is %s',
                img[0].data.dtype
            )
            img = promote_hdulist(img)
        if hasattr(self, 'run'):
            img = self.run(img)
        else:
            warnings.warn("use method 'run' instead of '_run'", DeprecationWarning,
                          stacklevel=2)
            img = self.run(img)

        return img

    def get_imgid(self, img, prefix=False):
        return dm.get_imgid(img, prefix=prefix)


TagOptionalCorrector = Corrector


class BadPixelCorrector(Corrector):
    """A Node that corrects a frame from bad pixels."""

    def __init__(self, badpixelmask, datamodel=None,
                 calibid='calibid-unknown',
                 dtype='float32', hwin=2, wwin=2):

        super(BadPixelCorrector, self).__init__(datamodel, calibid, dtype)

        self.bpm = badpixelmask
        self.hwin = hwin
        self.wwin = wwin

    def run(self, img):
        import numina.array.bpm as bpm
        imgid = self.get_imgid(img)

        _logger.debug('correcting bad pixel mask in %s', imgid)

        data = self.datamodel.get_data(img)
        newdata = bpm.process_bpm_median(
            data, self.bpm, hwin=self.hwin, wwin=self.wwin,
            fill=0, reuse_values=True
        )
        # newdata = array.fixpix(data, self.bpm)
        # FIXME: this breaks datamodel abstraction
        img['primary'].data = newdata
        hdr = img['primary'].header
        hdr['NUM-BPM'] = self.calibid
        hdr['history'] = f'BPM correction with {self.calibid}'
        t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
        hdr['history'] = f'BPM correction time {t_str}'
        return img


class BiasCorrector(Corrector):
    """A Node that corrects a frame from bias."""

    def __init__(self, biasmap, biasvar=None, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        self.update_variance = True if biasvar else False

        super(BiasCorrector, self).__init__(
            datamodel=datamodel, calibid=calibid, dtype=dtype)
        self.bias_stats = biasmap.mean()
        self.biasmap = biasmap
        self.biasvar = biasvar

    def run(self, img):

        imgid = self.get_imgid(img)

        _logger.debug('correcting bias in %s', imgid)
        _logger.debug('bias mean is %f', self.bias_stats)

        data = self.datamodel.get_data(img)
        data -= self.biasmap
        # FIXME
        img[0].data = data

        if self.update_variance:
            _logger.debug('update variance with bias')
            variance = self.datamodel.get_variance(img)
            variance += self.biasvar
            # FIXME
            img[1].data = variance
        hdr = img['primary'].header
        hdr['NUM-BS'] = self.calibid
        hdr['history'] = f'Bias correction with {self.calibid}'
        hdr['history'] = f'Bias image mean is {self.bias_stats}'
        t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
        hdr['history'] = f'Bias correction time {t_str}'
        return img


class DarkCorrector(Corrector):
    """A Node that corrects a frame from dark current."""

    def __init__(self, darkmap, darkvar=None, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        self.update_variance = False

        if darkvar:
            self.update_variance = True

        super(DarkCorrector, self).__init__(datamodel=datamodel,
                                            calibid=calibid,
                                            dtype=dtype)

        self.dark_stats = darkmap.mean()
        self.darkmap = darkmap
        self.darkvar = darkvar

    def run(self, img):

        etime = self.datamodel.get_darktime(img)

        data = self.datamodel.get_data(img)

        data -= self.darkmap * etime
        # FIXME
        img[0].data = data

        if self.update_variance:
            variance = self.datamodel.get_variance(img)
            variance += self.darkvar * etime * etime
            # FIXME
            img[1].data = variance
        hdr = img['primary'].header
        hdr['NUM-DK'] = self.calibid
        hdr['history'] = f'Dark correction with {self.calibid}'
        t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
        hdr['history'] = f'Dark correction time {t_str}'
        return img


class NonLinearityCorrector(Corrector):
    """A Node that corrects a frame from non-linearity."""

    def __init__(self, polynomial, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        self.update_variance = False

        super(NonLinearityCorrector, self).__init__(
            datamodel=datamodel, calibid=calibid, dtype=dtype
        )

        self.polynomial = polynomial

    def run(self, img):
        _logger.debug('correcting non linearity in %s', img)

        data = self.datamodel.get_data(img)

        data = array.correct_nonlinearity(data, self.polynomial,
                                          dtype=self.dtype)
        # FIXME
        img[0].data = data
        hdr = self.datamodel.get_header(img)
        hdr['NUM-LIN'] = self.calibid
        hdr['history'] = f'Non-linearity correction with {self.calibid}'
        t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
        hdr['history'] = f'Non-linearity correction time {t_str}'
        return img


class FlatFieldCorrector(Corrector):
    """A Node that corrects a frame from flat-field."""

    def __init__(self, flatdata, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        self.update_variance = False

        super(FlatFieldCorrector, self).__init__(
            datamodel=datamodel,
            calibid=calibid,
            dtype=dtype)

        self.flatdata = flatdata
        self.flat_stats = flatdata.mean()

    def run(self, img):
        imgid = self.get_imgid(img)

        _logger.debug('correcting flat in %s', imgid)
        _logger.debug('flat mean is %f', self.flat_stats)

        data = self.datamodel.get_data(img)
        data = array.correct_flatfield(data, self.flatdata, dtype=self.dtype)
        # FIXME
        img[0].data = data
        hdr = img['primary'].header
        hdr['NUM-FF'] = self.calibid
        hdr['history'] = f'Flat-field correction with {self.calibid}'
        t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
        hdr['history'] = f'Flat-field correction time {t_str}'
        hdr['history'] = f'Flat-field correction mean {self.flat_stats}'
        return img


class SkyCorrector(Corrector):
    """A Node that corrects a frame from sky."""

    def __init__(self, skydata, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        self.update_variance = False

        super(SkyCorrector, self).__init__(
            datamodel=datamodel,
            calibid=calibid,
            dtype=dtype)

        self.skydata = skydata
        self.calib_stats = skydata.mean()

    def run(self, img):
        imgid = self.get_imgid(img)

        if self.datamodel.do_sky_correction(img):
            _logger.debug('correcting sky in %s', imgid)
            _logger.debug('sky mean is %f', self.calib_stats)

            data = self.datamodel.get_data(img)

            data = array.correct_sky(data, self.skydata, dtype=self.dtype)

            # FIXME
            img[0].data = data
            hdr = img['primary'].header
            hdr['NUM-SK'] = self.calibid
            hdr['history'] = f'Sky subtraction with {self.calibid}'
            t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
            hdr['history'] = f'Sky subtraction time {t_str}'
            hdr['history'] = f'Sky subtraction mean {self.calib_stats}'
        else:
            _logger.debug('skip sky correction in %s', imgid)
        return img


class DivideByExposure(Corrector):
    """A Node that divides its input by exposure time."""

    def __init__(self, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        self.update_variance = False

        super(DivideByExposure, self).__init__(
            datamodel=datamodel,
            calibid=calibid,
            dtype=dtype
        )

    def run(self, img):
        imgid = self.get_imgid(img)
        hdr = self.datamodel.get_header(img)
        bunit = hdr['BUNIT']
        convert_to_s = False
        if bunit:
            if bunit.lower() == 'adu':
                convert_to_s = True
            elif bunit.lower() == 'adu/s':
                convert_to_s = False
            else:
                _logger.warning('Unrecognized value for BUNIT %s', bunit)
        if convert_to_s:
            etime = self.datamodel.get_exptime(img)
            _logger.debug('divide %s by exposure time %f', imgid, etime)

            img[0].data /= etime
            img[0].header['BUNIT'] = 'ADU/s'
            try:
                img['variance'].data /= etime ** 2
                img['variance'].header['BUNIT'] = 'ADU/s'
            except KeyError:
                pass

            hdr['NUM-EXP'] = etime
            hdr['history'] = f'Divided by exposure {etime}'
            t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
            hdr['history'] = f'Divided by exposure {t_str}'
        return img



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import astropy.units as u
import numpy as np

from .raise_valueerror import raise_ValueError


def simulate_constant_photlam(wmin, wmax, nphotons, wavelength_sampling, rng):
    """Simulate spectrum with constant flux (in PHOTLAM units).

    Parameters
    ----------
    wmin : `~astropy.units.Quantity`
        Minimum wavelength to be considered.
    wmax : `~astropy.units.Quantity`
        Maximum wavelength to be considered.
    nphotons : int
        Number of photons to be simulated.
    wavelength_sampling : str
        Method to sample the wavelength values. Two options are valid:
        - 'random': the wavelengt of each photon is randomly determined
          using the spectrum shape as the density probability function.
        - 'fixed': the wavelength of each photon is exactly determined
          using the spectrum shape as the density probability function.
    rng : `~numpy.random._generator.Generator`
        Random number generator.

    """

    if not isinstance(wmin, u.Quantity):
        raise_ValueError(f"Object 'wmin': {wmin} is not a Quantity instance")
    if not isinstance(wmax, u.Quantity):
        raise_ValueError(f"Object 'wmax': {wmax} is not a Quantity instance")
    if wmin.unit != wmax.unit:
        raise_ValueError(f"Different units used for 'wmin' and 'wmax': {wmin.unit}, {wmax.unit}.\n" +
                         "Employ the same unit to unambiguously define the output result.")

    if wavelength_sampling == 'random':
        simulated_wave = rng.uniform(low=wmin.value, high=wmax.value, size=nphotons)
    elif wavelength_sampling == 'fixed':
        simulated_wave = np.linspace(wmin.value, wmax.value, num=nphotons)
    else:
        simulated_wave = None  # avoid PyCharm warning
        raise_ValueError(f'Unexpected {wavelength_sampling=}')

    simulated_wave *= wmin.unit
    return simulated_wave



# Copyright 2008-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import numpy

from .datatype import DataType


class ArrayType(DataType):
    """A type of array."""

    def __init__(self, fmt='%.18e', default=None):
        super(ArrayType, self).__init__(ptype=numpy.ndarray, default=default)
        self.fmt = fmt

    def convert(self, obj):
        return self.convert_to_array(obj)

    def convert_to_array(self, obj):
        return numpy.array(obj)

    def _datatype_dump(self, obj, where):
        return dump_numpy_array(obj, where, self.fmt)

    def _datatype_load(self, obj):
        if isinstance(obj, str):
            # if is a string, it may be a pathname, try to load it

            # heuristics, by extension
            if obj.endswith('.csv'):
                # try to open as a CSV file
                res = numpy.loadtxt(obj, delimiter=',')
            else:
                res = numpy.loadtxt(obj)
        else:
            res = obj
        return res


class ArrayNType(ArrayType):
    def __init__(self, dimensions, default=None):
        super(ArrayNType, self).__init__(default=default)
        self.N = dimensions


def get_filename_obj(hdul, hint, ext):

    if isinstance(hint, str):
        filename = hint + ext
    elif callable(hint):
        filename = hint(hdul)
    else:
        raise ValueError('"hint" is neither string nor callable')
    return filename


def get_filename_numpy_array(obj, where):
    filename = get_filename_obj(obj, where, '.txt')
    return filename


def dump_numpy_array(obj, where, fmt='%.18e'):
    filename = get_filename_numpy_array(obj, where)
    numpy.savetxt(filename, obj, fmt=fmt)
    return filename



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Tests for RecipeInput and RecipeResult"""

from ..dataholders import Parameter
from ..recipeinout import RecipeInput


def create_input_class():

    class BB(RecipeInput):
        param1 = Parameter(1, 'something1')
        param2 = Parameter(2, 'something2')

        def mayfun(self):
            pass

    return BB


def test_class_construction():

    BB = create_input_class()

    assert issubclass(BB, RecipeInput)


def test_class_desc_access():

    BB = create_input_class()

    assert BB.param1 is getattr(BB, 'param1')
    assert BB.param2 is getattr(BB, 'param2')

    assert isinstance(BB.param1, Parameter)
    assert isinstance(BB.param2, Parameter)


def test_class_desc_set():

    BB = create_input_class()

    BB.param3 = Parameter(3, 'something3')

    assert BB.param3 is getattr(BB, 'param3')
    assert isinstance(BB.param3, Parameter)
    assert BB.param3 is BB.stored()['param3']
    assert BB.param3.dest == 'param3'


def test_class_destination_set():

    class BB(RecipeInput):
        param3h2hd = Parameter(1, 'something1', destination="param3")

    assert BB.param3 is getattr(BB, 'param3')
    assert isinstance(BB.param3, Parameter)
    assert BB.param3 is BB.stored()['param3']
    assert BB.param3.dest == 'param3'


def test_class_desc_stored():

    BB = create_input_class()

    stored = BB.stored()

    assert BB.param1 is stored['param1']
    assert BB.param2 is stored['param2']


def test_ins_desc_access():

    BB = create_input_class()

    bb = BB(param1=80)

    values = {'param2': 2, 'param1': 80}

    for key, val in values.items():
        ival = getattr(bb, key)
        assert val == ival

    # These values are not stored in __dict__
    for key in values:
        assert key not in bb.__dict__


def test_ins_attr_access():

    BB = create_input_class()

    bb = BB(param1=80)

    values = {'param2': 2, 'param1': 80}

    for key, val in bb.attrs().items():
        assert val == values[key]

    # These values are not stored in __dict__
    for key in values:
        assert key not in bb.__dict__


def test_class_nondesc_access():

    BB = create_input_class()
    bb = BB(param1=80)

    values = {'param2': 2, 'param1': 80}

    # If we insert a new attribute,it goes to __dict__
    bb.otherattr = 100

    for key, val in bb.attrs().items():
        assert val == values[key]

    assert 'otherattr' in bb.__dict__




import astropy.io.fits as fits
import pytest

from numina.instrument.assembly import assembly_instrument
from numina.instrument.collection import load_paths_store


def load_from_tests(name, date_str):
    # pkg_paths = ["pattridrp.instrument.configs"]
    pkg_paths = ["numina.drps.tests.configs"]
    comp_store = load_paths_store(pkg_paths)
    ins = assembly_instrument(comp_store, name, date_str)
    return ins


@pytest.fixture
def pattri_ins():
    ins = load_from_tests("PATTRI", "2016-06-01T12:00:00.0")
    return ins


@pytest.fixture
def pattri_header():
    hdr = fits.Header()
    hdr["FILTER"] = "R"  # filter -> PATTRI.wheel.filter -> FILTER
    hdr["tag0"] = "tag0_A"
    hdr["tag1"] = "tag1_X"
    hdr["insmode"] = "Mode_B"
    hdr["val"] = "A"
    hdr["R00_ANGL"] = 0.0
    hdr["R01_ANGL"] = -0.3
    hdr["R02_ANGL"] = 0.1
    hdr["R03_ANGL"] = 0.4
    hdr["R04_ANGL"] = 0.4
    hdr["R05_ANGL"] = 0.1
    hdr["R06_ANGL"] = -0.3
    hdr["R00_ACTV"] = True
    hdr["R01_ACTV"] = True
    hdr["R02_ACTV"] = True
    hdr["R03_ACTV"] = False
    hdr["R04_ACTV"] = True
    hdr["R05_ACTV"] = True
    hdr["R06_ACTV"] = True
    return hdr


@pytest.fixture
def pattri_header2():
    hdr = fits.Header()
    hdr["POS"] = 3
    hdr["FILTER"] = "R"
    hdr["tag0"] = "tag0_A"
    hdr["tag1"] = "tag1_X"
    hdr["insmode"] = "Mode_A"
    hdr["R00_ANGL"] = 0.0
    hdr["R01_ANGL"] = -0.3
    hdr["R02_ANGL"] = 0.1
    hdr["R03_ANGL"] = 0.4
    hdr["R04_ANGL"] = 0.4
    hdr["R05_ANGL"] = 0.1
    hdr["R06_ANGL"] = -0.3
    hdr["R00_ACTV"] = False
    hdr["R01_ACTV"] = False
    hdr["R02_ACTV"] = False
    hdr["R03_ACTV"] = False
    hdr["R04_ACTV"] = False
    hdr["R05_ACTV"] = False
    hdr["R06_ACTV"] = False
    return hdr


@pytest.fixture
def pattri_header_err():
    hdr = fits.Header()
    hdr["POS"] = 3
    hdr["tag1"] = 1
    hdr["tag0"] = 1
    hdr["insmode"] = "Mode_A"
    hdr["filter"] = "U"
    return hdr


@pytest.fixture
def pattri_state():
    state = dict()
    state["PATTRI"] = {}
    state["PATTRI.wheel"] = {"label": "U"}
    state["PATTRI.robot.arm_2"] = {"angle": 1.2, "active": False}
    state["PATTRI.PSU"] = {"mode": "Mode_B"}
    state["PATTRI.PSU.BarL_1"] = {"pos": -8}
    return state



#
# Copyright 2017-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import ast


def split_type_name(name):
    fp = name.find('(')
    sp = name.find(')')
    if (fp == -1) or (sp == -1) or (sp < fp):
        # paren not found
        klass = name
        fargs = ""
    else:
        klass = name[:fp]
        fargs = name[fp + 1:sp]
    return klass, fargs


def parse_arg_line(fargs):
    """parse limited form of arguments of function

    in the form a=1, b='c'
    as a dictionary
    """

    # Convert to literal dict
    fargs = fargs.strip()
    if fargs == '':
        return {}

    pairs = [s.strip() for s in fargs.split(',')]
    # find first "="
    result = []
    for p in pairs:
        fe = p.find("=")
        if fe == -1:
            # no equal
            raise ValueError("malformed")
        key = p[:fe]
        val = p[fe + 1:]
        tok = f"'{key}': {val}"
        result.append(tok)
    tokj = ','.join(result)
    result = f"{{ {tokj} }}"
    state = ast.literal_eval(result)
    return state



import numpy as np
from numina.array.distortion import shift_image2d


def test_simple_offset():
    a = np.zeros(25, dtype=float).reshape(5, 5)
    a_expected = np.zeros(25, dtype=float).reshape(5, 5)

    for i in range(1, 3):
        for j in range(1, 3):
            a[i, j] = 1

    a_expected[1, 1] = 0.25
    a_expected[1, 2] = 0.50
    a_expected[1, 3] = 0.25
    a_expected[2, 1] = 0.50
    a_expected[2, 2] = 1.00
    a_expected[2, 3] = 0.50
    a_expected[3, 1] = 0.25
    a_expected[3, 2] = 0.50
    a_expected[3, 3] = 0.25

    a_shifted = shift_image2d(a, xoffset=0.5, yoffset=0.5, resampling=2)

    assert np.allclose([a.sum()], [a_expected.sum()])
    assert np.allclose([a.sum()], [a_shifted.sum()])
    assert np.allclose(a_shifted, a_expected)


if __name__ == '__main__':
    test_simple_offset()



#
# Copyright 2017-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Convert strings to values in data load """

import datetime

import dateutil.parser


def convert_date(value):
    """
    Convert any string to datetime

    Parameters
    ----------
    value : str or datetime

    Returns
    -------

    """
    if isinstance(value, datetime.datetime):
        return value
    if value:
        return dateutil.parser.isoparse(value)
    else:
        return None


def convert_qc(value):
    """
    Convert string to QC

    Parameters
    ----------
    value : str

    Returns
    -------
    numina.types.qc.QC

    """
    from numina.types.qc import QC
    if value:
        if isinstance(value, QC):
            return value
        else:
            return QC[value]
    else:
        return QC.UNKNOWN
